import json
import os
import platform
import warnings
from os.path import dirname, join, realpath
from pathlib import Path
from typing import Any, Dict, Union

import pytest
import vcr
import yaml
from click.testing import CliRunner, Result
from pyfakefs.fake_filesystem import FakeFilesystem
from pygitguardian import GGClient
from pygitguardian.models import ScanResult, SecretIncident
from requests.utils import DEFAULT_CA_BUNDLE_PATH, extract_zipped_paths

from ggshield.core.cache import Cache
from ggshield.core.ui.reset import reset
from ggshield.core.url_utils import dashboard_to_api_url
from ggshield.utils.git_shell import (
    _get_git_path,
    _git_rev_parse_absolute,
    read_git_file,
)
from tests.conftest import GG_VALID_TOKEN


def is_macos():
    return platform.system() == "Darwin"


DATA_PATH = Path(__file__).parent.absolute() / "data"


_MULTIPLE_SECRETS_PATCH_CONTENT = """@@ -0,0 +1,2 @@
+FacebookAppKeys :
+String docker run --name geonetwork -d \
            -p 8080:8080 -e MYSQL_HOST=google.com \
            -e MYSQL_PORT=5434 -e MYSQL_USERNAME=root \
            -e MYSQL_PASSWORD=m42ploz2wd geonetwork
"""

_MULTIPLE_SECRETS_PATCH = (
    """commit 9537b6343a81f88d471e93f20ffb2e2665bbab00
Author: GitGuardian Owl <owl@example.com>
Date:   Thu Aug 18 18:20:21 2022 +0200

A message

:000000 100644 0000000 e965047 A\0test.txt\0\0diff --git a/test.txt b/test.txt
new file mode 100644
index 0000000..b80e3df
--- /dev/null
+++ b/test.txt
"""
    + _MULTIPLE_SECRETS_PATCH_CONTENT
)

_MULTIPLE_SECRETS_SCAN_RESULT = ScanResult.SCHEMA.load(
    {
        "policy_break_count": 1,
        "policies": ["Secrets detection", "File extensions", "Filenames"],
        "policy_breaks": [
            {
                "type": "MySQL Assignment",
                "policy": "Secrets detection",
                "matches": [
                    {
                        "type": "host",
                        "match": "google.com",
                        "index_start": 114,
                        "index_end": 123,
                        "line_start": 3,
                        "line_end": 3,
                    },
                    {
                        "type": "port",
                        "match": "5434",
                        "index_start": 151,
                        "index_end": 154,
                        "line_start": 3,
                        "line_end": 3,
                    },
                    {
                        "type": "username",
                        "match": "root",
                        "index_start": 174,
                        "index_end": 177,
                        "line_start": 3,
                        "line_end": 3,
                    },
                    {
                        "type": "password",
                        "match": "m42ploz2wd",
                        "index_start": 209,
                        "index_end": 218,
                        "line_start": 3,
                        "line_end": 3,
                    },
                ],
            }
        ],
    }
)

API_TOKENS_RESPONSE_SCAN_SCOPES = {
    "id": "5ddaad0c-5a0c-4674-beb5-1cd198d13360",
    "name": "myTokenName",
    "workspace_id": 42,
    "type": "personal_access_token",
    "status": "revoked",
    "created_at": "2023-05-20T12:40:55.662949Z",
    "last_used_at": "2023-05-24T12:40:55.662949Z",
    "expire_at": None,
    "revoked_at": "2023-05-27T12:40:55.662949Z",
    "member_id": 22015,
    "creator_id": 22015,
    "scopes": ["scan"],
}

# This long token is a test token, always reported as an uncheckable secret
GG_TEST_TOKEN = (
    "8a784aab7090f6a4ba3b9f7a6594e2e727007a26590b58ed314e4b9ed4536479sRZlRup3xvtMVfiHWA"
    "anbe712Jtc3nY8veZux5raL1bhpaxiv0rfyhFoAIMZUCh2Njyk7gRVsSQFPrEphSJnxa16SIdWKb03sRft"
    "770LUTTYTAy3IM18A7Su4HjiHlGA9ihLj9ou3luadfRAATlKH6kAZwTw289Kq9uip67zxyWkUJdh6PTeFp"
    "MgCh3AhHcZ21VeZHlu12345"
)

UNCHECKED_SECRET_PATCH = f"""commit 9537b6343a81f88d471e93f20ffb2e2665bbab00
Author: GitGuardian Owl <owl@example.com>
Date:   Thu Aug 18 18:20:21 2022 +0200

A message

:000000 100644 0000000 e965047 A\0test\0\0diff --git a/test b/test
new file mode 100644
index 0000000..b80e3df
--- /dev/null
+++ b/test
@@ -0,0 +2 @@
+# gg token
+apikey = "{GG_TEST_TOKEN}";
"""

VALID_SECRET_PATCH = f"""commit 9537b6343a81f88d471e93f20ffb2e2665bbab00
Author: GitGuardian Owl <owl@example.com>
Date:   Thu Aug 18 18:20:21 2022 +0200

A message

:000000 100644 0000000 e965047 A\0test\0\0diff --git a/test b/test
new file mode 100644
index 0000000..b80e3df
--- /dev/null
+++ b/test
@@ -0,0 +2 @@
+# gg token
+apikey = "{GG_VALID_TOKEN}";
"""

_SIMPLE_SECRET_TOKEN = "368ac3edf9e850d1c0ff9d6c526496f8237ddf91"  # ggignore
_SIMPLE_SECRET_PATCH = f"""@@ -0,0 +1 @@
+github_token: {_SIMPLE_SECRET_TOKEN}
"""
_SIMPLE_SECRET_PATCH_SCAN_RESULT = ScanResult.SCHEMA.load(
    {
        "policies": ["File extensions", "Filenames", "Secrets detection"],
        "policy_breaks": [
            {
                "type": "GitHub Token",
                "policy": "Secrets Detection",
                "matches": [
                    {
                        "match": _SIMPLE_SECRET_TOKEN,
                        "type": "apikey",
                        "index_start": 29,
                        "index_end": 69,
                    }
                ],
            }
        ],
        "policy_break_count": 1,
    }
)

_SIMPLE_SECRET_WITH_FILENAME_PATCH_SCAN_RESULT = ScanResult.SCHEMA.load(
    {
        "policies": ["File extensions", "Filenames", "Secrets detection"],
        "policy_breaks": [
            {
                "type": ".env",
                "policy": "Filenames",
                "matches": [{"type": "filename", "match": ".env"}],
            },
            {
                "type": "GitHub Token",
                "policy": "Secrets Detection",
                "matches": [
                    {
                        "match": _SIMPLE_SECRET_TOKEN,  # noqa
                        "type": "apikey",
                        "index_start": 29,
                        "index_end": 69,
                    }
                ],
            },
        ],
        "policy_break_count": 2,
    }
)

_MULTI_SECRET_ONE_LINE_PATCH = """@@ -0,0 +1 @@
+FacebookAppId = 294790898041575; FacebookAppSecret = ce3f9f0362bbe5ab01dfc8ee565e4372;

"""

_MULTI_SECRET_ONE_LINE_FULL_PATCH = (
    """commit 9537b6343a81f88d471e93f20ffb2e2665bbab00
Author: GitGuardian Owl <owl@example.com>
Date:   Thu Aug 18 18:20:21 2022 +0200

A message

:000000 100644 0000000 e965047 A\0test\0\0diff --git a/test b/test
new file mode 100644
index 0000000..3c9af3f
--- /dev/null
+++ b/test
"""
    + _MULTI_SECRET_ONE_LINE_PATCH
)

_MULTI_SECRET_ONE_LINE_PATCH_SCAN_RESULT = ScanResult.SCHEMA.load(
    {
        "policies": ["File extensions", "Filenames", "Secrets detection"],
        "policy_breaks": [
            {
                "type": "Facebook Access Tokens",
                "policy": "Secrets Detection",
                "matches": [
                    {
                        "match": "294790898041575",
                        "index_start": 31,
                        "index_end": 45,
                        "type": "client_id",
                    },
                    {
                        "match": "ce3f9f0362bbe5ab01dfc8ee565e4372",
                        "index_start": 68,
                        "index_end": 99,
                        "type": "client_secret",
                    },
                ],
            }
        ],
        "policy_break_count": 1,
    }
)


_MULTI_SECRET_ONE_LINE_PATCH_OVERLAY = """@@ -0,0 +1 @@
+Facebook = 294790898041575 | ce3f9f0362bbe5ab01dfc8ee565e4372;

"""

_MULTI_SECRET_ONE_LINE_PATCH_OVERLAY_SCAN_RESULT = ScanResult.SCHEMA.load(
    {
        "policies": ["File extensions", "Filenames", "Secrets detection"],
        "policy_breaks": [
            {
                "type": "Facebook Access Tokens",
                "policy": "Secrets Detection",
                "matches": [
                    {
                        "match": "294790898041575",
                        "index_start": 26,
                        "index_end": 40,
                        "type": "client_id",
                    },
                    {
                        "match": "ce3f9f0362bbe5ab01dfc8ee565e4372",
                        "index_start": 44,
                        "index_end": 75,
                        "type": "client_secret",
                    },
                ],
            }
        ],
        "policy_break_count": 1,
    }
)

_MULTI_SECRET_TWO_LINES_PATCH = """@@ -0,0 +2 @@
+FacebookAppId = 294790898041575;
+FacebookAppSecret = ce3f9f0362bbe5ab01dfc8ee565e4372;

"""

_MULTI_SECRET_TWO_LINES_PATCH_SCAN_RESULT = ScanResult.SCHEMA.load(
    {
        "policies": ["File extensions", "Filenames", "Secrets detection"],
        "policy_breaks": [
            {
                "type": "Facebook Access Tokens",
                "policy": "Secrets Detection",
                "matches": [
                    {
                        "match": "294790898041575",
                        "index_start": 31,
                        "index_end": 45,
                        "type": "client_id",
                    },
                    {
                        "match": "ce3f9f0362bbe5ab01dfc8ee565e4372",
                        "index_start": 69,
                        "index_end": 100,
                        "type": "client_secret",
                    },
                ],
            }
        ],
        "policy_break_count": 1,
    }
)

_MULTILINE_SECRET = """-----BEGIN RSA PRIVATE KEY-----
+MIIBOgIBAAJBAIIRkYjxjE3KIZiEc8k4sWWGNsPYRNE0u0bl5oFVApPLm+uXQ/4l
+bKO9LFtMiVPy700oMWLScwAN5OAiqVLMvHUCAwEAAQJANLr8nmEWuV6t2hAwhK5I
+NNmBkEo4M/xFxEtl9J7LKbE2gtNrlCQiJlPP1EMhwAjDOzQcJ3lgFB28dkqH5rMW
+TQIhANrCE7O+wlCKe0WJqQ3lYlHG91XWyGVgfExJwBDsAD9LAiEAmDY5OSsH0n2A
+22tthkAvcN1s66lG+0DztOVJ4QLI2z8CIBPeDGwGpx8pdIicN/5LFuLWbyAcoZaT
+bLaA/DCNPniBAiA0l//bzg+M3srIhm04xzLdR9Vb9IjPRlkvN074zdKDVwIhAKJb
+RF3C+CMFb0wXme/ovcDeM1+3W/UmSHYUW4b3WYq4
+-----END RSA PRIVATE KEY-----"""

_SIMPLE_SECRET_MULTILINE_PATCH_SCAN_RESULT = ScanResult.SCHEMA.load(
    {
        "policies": ["File extensions", "Filenames", "Secrets detection"],
        "policy_breaks": [
            {
                "type": "RSA Private Key",
                "policy": "Secrets Detection",
                "matches": [
                    {
                        "match": _MULTILINE_SECRET,  # noqa
                        "index_start": 42,
                        "index_end": 543,
                        "type": "apikey",
                    }
                ],
            }
        ],
        "policy_break_count": 1,
    }
)

_SIMPLE_SECRET_MULTILINE_PATCH = (
    """@@ -0,0 +1,29 @@
+PrivateKeyRsa:
+- text: """
    + _MULTILINE_SECRET
)


_ONE_LINE_AND_MULTILINE_PATCH_SCAN_RESULT = ScanResult.SCHEMA.load(
    {
        "policy_breaks": [
            {
                "type": "Facebook Access Tokens",
                "policy": "Secrets Detection",
                "matches": [
                    {
                        "match": "294790898041573",
                        "line_start": 2,
                        "line_end": 2,
                        "index_start": 35,
                        "index_end": 49,
                        "type": "client_id",
                    },
                    {
                        "match": "ce3f9f0362bbe5ab01dfc8ee565e4371",
                        "line_start": 2,
                        "line_end": 2,
                        "index_start": 53,
                        "index_end": 84,
                        "type": "client_secret",
                    },
                ],
            },
            {
                "type": "RSA Private Key",
                "policy": "Secrets detection",
                "matches": [
                    {
                        "line_start": 2,
                        "match": _MULTILINE_SECRET,
                        "index_start": 86,
                        "index_end": 585,
                        "type": "apikey",
                        "line_end": 10,
                    }
                ],
            },
            {
                "type": "SendGrid Key",
                "policy": "Secrets detection",
                "matches": [
                    {
                        "line_start": 10,
                        "match": "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",  # noqa
                        "index_start": 594,
                        "index_end": 662,
                        "type": "apikey",
                        "line_end": 10,
                    }
                ],
            },
        ],
        "policies": ["Filenames", "File extensions", "Secrets detection"],
        "policy_break_count": 2,
    }
)

_ONE_LINE_AND_MULTILINE_PATCH_CONTENT = """@@ -0,0 +1,29 @@
+FacebookAppKeys: 294790898041573 / ce3f9f0362bbe5ab01dfc8ee565e4371 -----BEGIN RSA PRIVATE KEY-----
+MIIBOgIBAAJBAIIRkYjxjE3KIZiEc8k4sWWGNsPYRNE0u0bl5oFVApPLm+uXQ/4l
+bKO9LFtMiVPy700oMWLScwAN5OAiqVLMvHUCAwEAAQJANLr8nmEWuV6t2hAwhK5I
+NNmBkEo4M/xFxEtl9J7LKbE2gtNrlCQiJlPP1EMhwAjDOzQcJ3lgFB28dkqH5rMW
+TQIhANrCE7O+wlCKe0WJqQ3lYlHG91XWyGVgfExJwBDsAD9LAiEAmDY5OSsH0n2A
+22tthkAvcN1s66lG+0DztOVJ4QLI2z8CIBPeDGwGpx8pdIicN/5LFuLWbyAcoZaT
+bLaA/DCNPniBAiA0l//bzg+M3srIhm04xzLdR9Vb9IjPRlkvN074zdKDVwIhAKJb
+RF3C+CMFb0wXme/ovcDeM1+3W/UmSHYUW4b3WYq4
+-----END RSA PRIVATE KEY----- token: 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
"""  # noqa

_ONE_LINE_AND_MULTILINE_PATCH = (
    """commit 9537b6343a81f88d471e93f20ffb2e2665bbab00
Author: GitGuardian Owl <owl@example.com>
Date:   Thu Aug 18 18:20:21 2022 +0200

A message

:000000 100644 0000000 e965047 A\0test\0\0diff --git a/test b/test
new file mode 100644
index 0000000..b80e3df
--- /dev/null
+++ b/test
"""
    + _ONE_LINE_AND_MULTILINE_PATCH_CONTENT
)

_NO_SECRET_PATCH = """commit 9537b6343a81f88d471e93f20ffb2e2665bbab00
Author: GitGuardian Owl <owl@example.com>
Date:   Thu Aug 18 18:20:21 2022 +0200

A message

:000000 100644 0000000 e965047 A\0test\0\0diff --git a/test b/test
new file mode 100644
index 0000000..b80e3df
--- /dev/null
+++ b/test
@@ -0,0 +1 @@
+this is a patch without secret
"""

"""
Indices for _SECRET_RAW_FILE
                    0         1         2         3         4         5         6         7         8
                    012345678901234567890123456789012345678901234567890123456789012345678901234567890
"""
_SECRET_RAW_FILE = '+sg_key = "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";\n'

_SINGLE_ADD_PATCH = """commit 9537b6343a81f88d471e93f20ffb2e2665bbab00
Author: GitGuardian Owl <owl@example.com>
Date:   Thu Aug 18 18:20:21 2022 +0200

A message

:000000 100644 0000000 e965047 A\0test\0\0diff --git a/test b/test
new file mode 100644
index 0000000..3c9af3f
--- /dev/null
+++ b/test
@@ -0,0 +1 @@
+sg_key = "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";
"""
"""
Indices for _SINGLE_ADD_PATCH.
0         1         2         3         4         5         6         7         8
012345678901234567890123456789012345678901234567890123456789012345678901234567890
"""

_SINGLE_MOVE_PATCH = """
commit 9537b6343a81f88d471e93f20ffb2e2665bbab00
Author: GitGuardian Owl <owl@example.com>
Date:   Thu Aug 18 18:20:21 2022 +0200

A message

:000000 100644 0000000 e965047 M\0test\0\0diff --git a/test b/test
index 3c9af3f..b0ce1c7 100644
--- a/test
+++ b/test
@@ -150 +150,2 @@
+something
 sg_key = "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";
"""

_SINGLE_DELETE_PATCH = """commit 9537b6343a81f88d471e93f20ffb2e2665bbab00
Author: GitGuardian Owl <owl@example.com>
Date:   Thu Aug 18 18:20:21 2022 +0200

A message

:000000 100644 0000000 e965047 D\0test\0\0diff --git a/test b/test
index b0ce1c7..deba01f 100644
--- a/test
+++ b/test
@@ -1,2 +1 @@
 something
-sg_key = "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";
"""

TWO_POLICY_BREAKS = ScanResult.SCHEMA.load(
    {
        "policy_breaks": [
            {
                "type": "RSA Private Key",
                "policy": "Secrets detection",
                "matches": [
                    {
                        "line_start": 2,
                        "match": _MULTILINE_SECRET,
                        "index_start": 86,
                        "index_end": 585,
                        "type": "apikey",
                        "line_end": 10,
                    }
                ],
            },
            {
                "type": "SendGrid Key",
                "policy": "Secrets detection",
                "matches": [
                    {
                        "line_start": 10,
                        "match": "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",  # noqa
                        "index_start": 594,
                        "index_end": 662,
                        "type": "apikey",
                        "line_end": 10,
                    }
                ],
            },
        ],
        "policies": ["Filenames", "File extensions", "Secrets detection"],
        "policy_break_count": 2,
    }
)
_SINGLE_LINE_SECRET_FILE = f"""\
apikey = "{GG_TEST_TOKEN}";
"""

_MULTILINE_SECRET_FILE = """-----BEGIN RSA PRIVATE KEY-----
MIIBOgIBAAJBAIIRkYjxjE3KIZiEc8k4sWWGNsPYRNE0u0bl5oFVApPLm+uXQ/4l
bKO9LFtMiVPy700oMWLScwAN5OAiqVLMvHUCAwEAAQJANLr8nmEWuV6t2hAwhK5I
NNmBkEo4M/xFxEtl9J7LKbE2gtNrlCQiJlPP1EMhwAjDOzQcJ3lgFB28dkqH5rMW
TQIhANrCE7O+wlCKe0WJqQ3lYlHG91XWyGVgfExJwBDsAD9LAiEAmDY5OSsH0n2A
22tthkAvcN1s66lG+0DztOVJ4QLI2z8CIBPeDGwGpx8pdIicN/5LFuLWbyAcoZaT
bLaA/DCNPniBAiA0l//bzg+M3srIhm04xzLdR9Vb9IjPRlkvN074zdKDVwIhAKJb
RF3C+CMFb0wXme/ovcDeM1+3W/UmSHYUW4b3WYq4
-----END RSA PRIVATE KEY-----""".replace(
    "\r\n", "\n"
)

_ONE_LINE_AND_MULTILINE_FILE = """
FacebookAppKeys: 294790898041573 / ce3f9f0362bbe5ab01dfc8ee565e4371 -----BEGIN RSA PRIVATE KEY-----
MIIBOgIBAAJBAIIRkYjxjE3KIZiEc8k4sWWGNsPYRNE0u0bl5oFVApPLm+uXQ/4l
bKO9LFtMiVPy700oMWLScwAN5OAiqVLMvHUCAwEAAQJANLr8nmEWuV6t2hAwhK5I
NNmBkEo4M/xFxEtl9J7LKbE2gtNrlCQiJlPP1EMhwAjDOzQcJ3lgFB28dkqH5rMW
TQIhANrCE7O+wlCKe0WJqQ3lYlHG91XWyGVgfExJwBDsAD9LAiEAmDY5OSsH0n2A
22tthkAvcN1s66lG+0DztOVJ4QLI2z8CIBPeDGwGpx8pdIicN/5LFuLWbyAcoZaT
bLaA/DCNPniBAiA0l//bzg+M3srIhm04xzLdR9Vb9IjPRlkvN074zdKDVwIhAKJb
RF3C+CMFb0wXme/ovcDeM1+3W/UmSHYUW4b3WYq4
-----END RSA PRIVATE KEY----- token: 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
"""

# Docker example constants
DOCKER_EXAMPLE_PATH = DATA_PATH / "docker-example.tar.xz"
DOCKER__INCOMPLETE_MANIFEST_EXAMPLE_PATH = (
    DATA_PATH / "docker-incomplete-manifest-example.tar.xz"
)

# Format is { layer_id: { path: content }}
DOCKER_EXAMPLE_LAYER_FILES = {
    "sha256:4e850fb0fe03eae7a9a505d114b342ecdf7fd6e5a3ed2a1967e40083d63c7abe": {
        "/app/file_one": "Hello, I am the first file!\n"
    },
    "sha256:f1c86b269b6a35c2158e8cd69b5e276d6c238272adcdd44efa21e5d79d099ddb": {
        "/app/file_three.sh": "echo Life is beautiful.\n",
        "/app/file_two.py": """print("Hi! I'm the second file but I'm happy.")\n""",
    },
}


def _filter_vcr_response(response: Dict[str, Any]) -> Dict[str, Any]:
    # Read content-type
    headers = response["headers"]
    for name in ("content-type", "Content-Type"):
        try:
            content_type = headers[name][0]
            break
        except KeyError:
            pass
    else:
        # No content-type, let's assume it's not JSON
        return response

    # Remove token returned by calls to https://api.gitguardian.com/v1/auth/jwt.
    # Unfortunately we don't have the URL of the request at this point. We only remove
    # the token if it's the only field. If we don't we break honeytoken cassettes.
    if content_type == "application/json":
        content = json.loads(response["body"]["string"])
        if isinstance(content, dict) and list(content.keys()) == ["token"]:
            content["token"] = "<REDACTED>"
            response["body"]["string"] = json.dumps(content).encode()
    return response


my_vcr = vcr.VCR(
    cassette_library_dir=join(dirname(realpath(__file__)), "cassettes"),
    path_transformer=vcr.VCR.ensure_suffix(".yaml"),
    decode_compressed_response=True,
    ignore_localhost=True,
    match_on=["method", "url"],
    serializer="yaml",
    record_mode="once",
    filter_headers=["Authorization"],
    before_record_response=_filter_vcr_response,
)


@pytest.fixture(scope="session", autouse=True)
def _ensure_gitguardian_api_key():
    """
    Many VCR-based tests expect GITGUARDIAN_API_KEY to be set, set to a dummy one if
    it's not
    """
    if "GITGUARDIAN_API_KEY" not in os.environ:
        warnings.warn(
            "GITGUARDIAN_API_KEY is not set, recording VCR cassettes won't work."
        )
        os.environ["GITGUARDIAN_API_KEY"] = "not-a-real-key"


@pytest.fixture(scope="session")
def client(_ensure_gitguardian_api_key) -> GGClient:
    api_key = os.environ["GITGUARDIAN_API_KEY"]

    if "GITGUARDIAN_API_URL" in os.environ:  # deprecated
        base_uri = os.environ["GITGUARDIAN_API_URL"]
    else:
        instance_url = os.getenv(
            "GITGUARDIAN_INSTANCE", "https://dashboard.gitguardian.com"
        )
        base_uri = dashboard_to_api_url(instance_url)

    return GGClient(api_key, base_uri)


@pytest.fixture
def no_api_key(monkeypatch):
    """Remove GITGUARDIAN_API_KEY from the environment, useful to test anonymous use"""
    monkeypatch.delenv("GITGUARDIAN_API_KEY", raising=False)


@pytest.fixture(scope="session")
def cache() -> Cache:
    c = Cache()
    c.purge()
    return c


@pytest.fixture()
def cli_runner():
    return CliRunner()


@pytest.fixture(scope="function")
def cli_fs_runner(cli_runner):
    with cli_runner.isolated_filesystem():
        yield cli_runner


@pytest.fixture(scope="function")
def isolated_fs(fs):
    # isolate fs but include CA bundle for https validation
    fs.add_real_directory(os.path.dirname(extract_zipped_paths(DEFAULT_CA_BUNDLE_PATH)))
    # add cassettes dir
    cassettes_dir = join(dirname(realpath(__file__)), "cassettes")
    fs.add_real_directory(cassettes_dir)
    # Add a fake OS-release file. It describes a linux OS
    mock_contents = """ID="ubuntu"\nVERSION_ID="22.04"\n"""
    f = fs.create_file("/etc/os-release")
    f.set_contents(mock_contents)


def write_text(filename: Union[str, Path], content: str):
    """Create a text file named `filename` with content `content.
    Create any missing dirs if necessary.

    Note that using `write_bytes(content.encode())` ensures the
    same content is created, independently of the OS
    (whereas using Path.write_text creates different line ends depending on the OS)
    """
    path = Path(filename)
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_bytes(content.encode())


def write_yaml(filename: Union[str, Path], data: Any):
    """Save data as a YAML file in `filename`, using `write_text()`"""
    write_text(filename, yaml.dump(data))


def assert_invoke_exited_with(result: Result, exit_code: int):
    msg = f"""
    Expected code {exit_code}, got {result.exit_code}.

    stdout:
    {result.stdout}

    stderr:
    {result.stderr if result.stderr_bytes is not None else ""}
    """
    assert result.exit_code == exit_code, msg


def assert_invoke_ok(result: Result):
    assert_invoke_exited_with(result, 0)


def make_fake_path_inaccessible(fs: FakeFilesystem, path: Union[str, Path]):
    """
    Make `path` inaccessible inside `fs`. This is useful to test IO permission errors.
    """

    # `force_unix_mode` is required for Windows.
    # See <https://pytest-pyfakefs.readthedocs.io/en/latest/usage.html#set-file-as-inaccessible-under-windows>
    fs.chmod(path, 0o0000, force_unix_mode=True)


@pytest.fixture(autouse=True)
def clear_cache():
    _get_git_path.cache_clear()
    _git_rev_parse_absolute.cache_clear()
    read_git_file.cache_clear()


@pytest.fixture(autouse=True)
def _reset_ui_fixture():
    """
    Enabling debug mode has global side effects. Reset it to ensure a test touching the
    log or debug configuration does not affect other tests.
    """
    reset()


SECRET_INCIDENT_MOCK = SecretIncident.from_dict(
    {
        "id": 42,
        "date": "2024-09-04T16:00:48.956235+00:00",
        "detector": {
            "name": "detector",
            "display_name": "Detector",
            "nature": "specific",
            "family": "Api",
            "detector_group_name": "detector_group",
            "detector_group_display_name": "Detector Group",
        },
        "secret_hash": "NpQakTjuW7LlWWgeHOR5VewWfLbUtKn1bZ3EDrLLw7aMf26zyCrzdQOwnLqJOGTb",
        "hmsl_hash": "0483663d03dc4ea2c2bfb73d0a4a8bc6f2035d911f3a7d210a16e8314413c29e",
        "gitguardian_url": "https://dashboard.staging.gitguardian.tech/workspace/1/incidents/42",
        "regression": False,
        "status": "TRIGGERED",
        "assignee_id": None,
        "assignee_email": None,
        "occurrences_count": 1,
        "secret_presence": {
            "files_requiring_code_fix": 1,
            "files_pending_merge": 0,
            "files_fixed": 0,
            "outside_vcs": 0,
            "removed_outside_vcs": 0,
            "in_vcs": 1,
            "removed_in_vcs": 0,
        },
        "ignore_reason": None,
        "triggered_at": "2024-09-04T16:00:48.956235+00:00",
        "ignored_at": None,
        "ignorer_id": None,
        "ignorer_api_token_id": None,
        "resolver_id": None,
        "resolver_api_token_id": None,
        "secret_revoked": False,
        "severity": "high",
        "validity": "invalid",
        "resolved_at": None,
        "share_url": None,
        "tags": [],
        "feedback_list": [],
        "occurrences": [],
    }
)



from __future__ import annotations

import json
import re
import subprocess
import tarfile
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING

from click import UsageError

from ggshield.core import ui
from ggshield.core.dirs import get_cache_dir
from ggshield.core.errors import UnexpectedError
from ggshield.core.scan import Scannable, StringScannable
from ggshield.core.scan.id_cache import IDCache
from ggshield.core.scanner_ui import create_scanner_ui
from ggshield.utils.files import is_path_binary

from .secret_scan_collection import SecretScanCollection
from .secret_scanner import SecretScanner


if TYPE_CHECKING:
    from typing import Any, Dict, Generator, Iterable, List, Set

    from pygitguardian import GGClient

    from ggshield.core.cache import Cache
    from ggshield.core.config.user_config import SecretConfig
    from ggshield.core.scan import ScanContext

FILEPATH_BANLIST = [
    r"^/?usr/(?!share/nginx)",
    r"^/?lib/",
    r"^/?share/",
    r"^/?bin/",
    r"^/?sbin/",
    r"^/?node_modules/",
    r"^/?include/",
    r"^/?vendor/",
    r"^/?texlive/",
    r"^/?var/",
    r"^/?fonts/",
    r"^/?npm/",
    r"^/?site-packages/",
]
FILEPATH_BANLIST_PATTERNS = {
    re.compile(banned_filepath) for banned_filepath in FILEPATH_BANLIST
}
# Note that FILEPATH_BANLIST_PATTERNS comes in addition to what's done in _exclude_callback

LAYER_TO_SCAN_PATTERN = re.compile(r"\b(copy|add)\b", re.IGNORECASE)

TAG_PATTERN = re.compile(r":[a-zA-Z0-9_][-.a-zA-Z0-9_]{0,127}$")


class InvalidDockerArchiveException(Exception):
    pass


class DockerContentScannable(Scannable):
    """
    A Scannable for a file inside a Docker image
    """

    def __init__(
        self, layer_id: str, tar_file: tarfile.TarFile, tar_info: tarfile.TarInfo
    ):
        super().__init__()
        self._layer_id = layer_id
        self._tar_file = tar_file
        self._tar_info = tar_info

    @property
    def url(self) -> str:
        return f"{self._layer_id}:/{self._tar_info.name}"

    @property
    def filename(self) -> str:
        return self.url

    @property
    def path(self) -> Path:
        return Path("/", self._tar_info.name)

    def is_longer_than(self, max_utf8_encoded_size: int) -> bool:
        if self._utf8_encoded_size is not None:
            # We already have the encoded size, easy
            return self._utf8_encoded_size > max_utf8_encoded_size

        # We need to decode at least the beginning of the file to determine if it's
        # small enough
        fp = self._tar_file.extractfile(self._tar_info)
        assert fp is not None
        with fp:
            (
                result,
                self._content,
                self._utf8_encoded_size,
            ) = Scannable._is_file_longer_than(
                fp, max_utf8_encoded_size  # type:ignore
            )
            # mypy complains that fp is IO[bytes] but _is_file_longer_than() expects
            # BinaryIO. They are compatible, ignore the error.
        return result

    def _read_content(self) -> None:
        if self._content is None:
            file = self._tar_file.extractfile(self._tar_info)
            assert file is not None
            byte_content = file.read()
            self._content, self._utf8_encoded_size = Scannable._decode_bytes(
                byte_content
            )


@dataclass
class LayerInfo:
    filename: str
    command: str
    diff_id: str

    def should_scan(self) -> bool:
        """
        Returns True if a layer should be scanned, False otherwise.
        Only COPY and ADD layers should be scanned.
        """
        if self.command == "":
            # Some images contain layers with no commands. Since we don't know how they have
            # been created, we must scan them.
            # Examples of such images from Docker Hub:
            # - aevea/release-notary:0.9.7
            # - redhat/ubi8:8.6-754
            return True
        else:
            return LAYER_TO_SCAN_PATTERN.search(self.command) is not None


class DockerImage:
    # The manifest.json file
    manifest: Dict[str, Any]

    # The Image JSON file
    # (see https://github.com/moby/moby/blob/master/image/spec/v1.2.md#terminology)
    image: Dict[str, Any]

    layer_infos: List[LayerInfo]

    @staticmethod
    @contextmanager
    def open(archive_path: Path) -> Generator["DockerImage", None, None]:
        """ContextManager to create a DockerImage instance."""
        with tarfile.open(archive_path) as tar_file:
            yield DockerImage(archive_path, tar_file)

    def __init__(self, archive_path: Path, tar_file: tarfile.TarFile):
        """Creates a DockerImage instance. Internal. Prefer using DockerImage.open()."""
        self.archive_path = archive_path
        self.tar_file = tar_file
        self._load_manifest()
        self._load_image()
        self.config_scannable = StringScannable(
            "Dockerfile or build-args", json.dumps(self.image, indent=2)
        )

        self._load_layer_infos()

    def _load_manifest(self) -> None:
        """
        Reads "manifest.json", stores result in self.manifest
        """
        manifest_file = self.tar_file.extractfile("manifest.json")
        if manifest_file is None:
            raise InvalidDockerArchiveException("No manifest file found.")

        self.manifest = json.load(manifest_file)[0]

    def _load_image(self) -> None:
        """
        Reads the image JSON file, stores result in self.image
        """
        try:
            config_file_path = self.manifest["Config"]
        except KeyError:
            raise InvalidDockerArchiveException("No Config key in manifest.")

        config_file_info = self.tar_file.getmember(config_file_path)
        if config_file_info is None:
            raise InvalidDockerArchiveException("No config file found.")

        config_file = self.tar_file.extractfile(config_file_info)
        if config_file is None:
            raise InvalidDockerArchiveException("Config file could not be extracted.")

        self.image = json.load(config_file)

    def _load_layer_infos(self) -> None:
        """
        Fill self.layer_infos with LayerInfo instances for all non-empty layers
        """

        #
        # manifest["Layers"] contains a list of non-empty layers like this:
        #
        #   "<random_id>/layer.tar"
        layer_filenames = self.manifest["Layers"]

        # image["history"] contains a list of entries like this:
        # {
        #   "created": ISO8601 timestamp,
        #   "created_by": command to build the layer
        #   "empty_layer": if present, equals true
        # }
        non_empty_history_entries = [
            x for x in self.image["history"] if not x.get("empty_layer")
        ]

        #
        # image["rootfs"]["diff_ids"] contains the list of layer IDs
        diff_ids = self.image["rootfs"]["diff_ids"]

        layer_infos = [
            LayerInfo(
                filename=filename,
                command=history.get("created_by", ""),
                diff_id=diff_id,
            )
            for filename, history, diff_id in zip(
                layer_filenames, non_empty_history_entries, diff_ids
            )
        ]
        self.layer_infos = [x for x in layer_infos if x.should_scan()]

    def get_layer_scannables(
        self, layer_info: LayerInfo, exclusion_regexes: Set[re.Pattern[str]]
    ) -> Iterable[Scannable]:
        """
        Extracts Scannable to be scanned for given layer.
        """
        layer_filename = layer_info.filename
        layer_archive = tarfile.TarFile(
            name=self.archive_path / layer_filename,
            fileobj=self.tar_file.extractfile(layer_filename),
        )

        layer_id = layer_info.diff_id

        for file_info in layer_archive:
            if not file_info.isfile():
                continue

            if file_info.size == 0:
                continue

            if not _validate_filepath(
                filepath=file_info.path, exclusion_regexes=exclusion_regexes
            ):
                continue

            yield DockerContentScannable(layer_id, layer_archive, file_info)


def _validate_filepath(
    filepath: str,
    exclusion_regexes: Set[re.Pattern[str]],
) -> bool:
    if any(
        banned_pattern.search(filepath)
        for banned_pattern in FILEPATH_BANLIST_PATTERNS | exclusion_regexes
    ):
        return False

    if is_path_binary(filepath):
        return False
    return True


def _get_layer_id_cache(secrets_engine_version: str) -> IDCache:
    cache_path = get_cache_dir() / "docker" / f"{secrets_engine_version}.json"
    return IDCache(cache_path)


def docker_pull_image(image_name: str, timeout: int) -> None:
    """
    Pull docker image and raise exception on timeout or failed to find image

    Timeout after `timeout` seconds.
    """
    command = ["docker", "pull", image_name]
    try:
        subprocess.run(
            command,
            check=True,
            timeout=timeout,
        )
    except subprocess.CalledProcessError:
        raise UsageError(f'Image "{image_name}" not found')
    except subprocess.TimeoutExpired:
        raise UnexpectedError('Command "{}" timed out'.format(" ".join(command)))


def docker_save_to_tmp(image_name: str, destination_path: Path, timeout: int) -> None:
    """
    Do a `docker save <image_name> -o <destination_path>`

    Limit docker commands to run at most `timeout` seconds.
    """
    image_name = (
        image_name if TAG_PATTERN.search(image_name) else image_name + ":latest"
    )
    command = ["docker", "save", image_name, "-o", str(destination_path)]

    try:
        ui.display_info("Saving docker image...")
        subprocess.run(
            command,
            check=True,
            stderr=subprocess.PIPE,
            timeout=timeout,
        )
        ui.display_info("OK")
    except subprocess.CalledProcessError as exc:
        err_string = str(exc.stderr)
        if "No such image" in err_string or "reference does not exist" in err_string:
            ui.display_info("need to download image first")
            docker_pull_image(image_name, timeout)

            docker_save_to_tmp(image_name, destination_path, timeout)
        else:
            raise UnexpectedError(
                f"Unable to save docker archive:\nError: {err_string}"
            )
    except subprocess.TimeoutExpired:
        raise UnexpectedError('Command "{}" timed out'.format(" ".join(command)))


def docker_scan_archive(
    archive_path: Path,
    client: GGClient,
    cache: Cache,
    secret_config: SecretConfig,
    scan_context: ScanContext,
    exclusion_regexes: Set[re.Pattern[str]],
) -> SecretScanCollection:
    scanner = SecretScanner(
        client=client,
        cache=cache,
        scan_context=scan_context,
        secret_config=secret_config,
    )
    secrets_engine_version = client.secrets_engine_version
    assert secrets_engine_version is not None
    layer_id_cache = _get_layer_id_cache(secrets_engine_version)

    with DockerImage.open(archive_path) as docker_image:
        ui.display_heading("Scanning Docker config")
        with create_scanner_ui(1) as scanner_ui:
            results = scanner.scan(
                [docker_image.config_scannable], scanner_ui=scanner_ui
            )

        for info in docker_image.layer_infos:
            files = list(docker_image.get_layer_scannables(info, exclusion_regexes))
            file_count = len(files)
            if file_count == 0:
                continue
            print()
            layer_id = info.diff_id
            if layer_id in layer_id_cache:
                ui.display_heading(f"Skipping layer {layer_id}: already scanned")
            else:
                ui.display_heading(f"Scanning layer {info.diff_id}")
                with create_scanner_ui(file_count) as scanner_ui:
                    layer_results = scanner.scan(files, scanner_ui=scanner_ui)
                if not layer_results.has_secrets:
                    layer_id_cache.add(layer_id)
                results.extend(layer_results)

    return SecretScanCollection(
        id=str(archive_path), type="scan_docker_archive", results=results
    )



from pathlib import Path
from typing import Any

import click
from click import UsageError

from ggshield.cmd.secret.scan.secret_scan_common_options import (
    add_secret_scan_common_options,
    create_output_handler,
)
from ggshield.cmd.utils.common_decorators import exception_wrapper
from ggshield.cmd.utils.context_obj import ContextObj
from ggshield.core import ui
from ggshield.core.client import create_client_from_config
from ggshield.core.scan import ScanContext, ScanMode
from ggshield.utils.git_shell import get_list_commit_SHA
from ggshield.verticals.secret.repo import scan_commit_range


@click.command()
@click.argument("commit_range", nargs=1, type=click.STRING)
@add_secret_scan_common_options()
@click.pass_context
@exception_wrapper
def range_cmd(
    ctx: click.Context,
    commit_range: str,
    **kwargs: Any,
) -> int:  # pragma: no cover
    """
    Scan each commit in the given commit range.

    Any git compatible commit range can be provided as an input.

    Example: `ggshield secret scan commit-range HEAD~1...`
    """
    ctx_obj = ContextObj.get(ctx)
    ctx_obj.client = create_client_from_config(ctx_obj.config)
    config = ctx_obj.config
    commit_list = get_list_commit_SHA(commit_range)
    if not commit_list:
        raise UsageError("invalid commit range")
    ui.display_verbose(f"Commits to scan: {len(commit_list)}")

    scan_context = ScanContext(
        scan_mode=ScanMode.COMMIT_RANGE,
        command_path=ctx.command_path,
        target_path=Path.cwd(),
    )

    return scan_commit_range(
        client=ctx_obj.client,
        cache=ctx_obj.cache,
        commit_list=commit_list,
        output_handler=create_output_handler(ctx),
        exclusion_regexes=ctx_obj.exclusion_regexes,
        secret_config=config.user_config.secret,
        scan_context=scan_context,
    )



import subprocess
from pathlib import Path

import pytest

from ggshield.__main__ import cli
from ggshield.cmd.secret.scan.precommit import (
    check_is_merge_with_conflict,
    check_is_merge_without_conflict,
    get_merge_branch_from_reflog,
)
from ggshield.utils.os import cd
from tests.repository import Repository
from tests.unit.conftest import assert_invoke_ok, my_vcr


def test_is_merge_not_merge(tmp_path):
    """
    GIVEN  a commit that is not a merge
    WHEN check_is_merge_with_conflict() or check_is_merge_without_conflict() is called
    THEN they return False
    """
    repo = Repository.create(tmp_path, initial_branch="master")

    Path(tmp_path / "inital.md").write_text("Initial")
    repo.add(".")
    repo.create_commit("Initial commit on master")

    assert not check_is_merge_without_conflict()
    assert not check_is_merge_with_conflict(cwd=tmp_path)


def test_is_merge_without_conflict(tmp_path, monkeypatch):
    """
    GIVEN  a merge commit without conflict
    WHEN check_is_merge_with_conflict() or check_is_merge_without_conflict() are called
    THEN they return the expected results
    """
    repo = Repository.create(tmp_path, initial_branch="master")
    repo.create_commit("Initial commit on master")

    repo.create_branch("feature_branch")
    repo.checkout("master")
    Path(tmp_path / "Other.md").write_text("Other")
    repo.add(".")
    repo.create_commit("Commit on master")

    repo.checkout("feature_branch")
    Path(tmp_path / "Another.md").write_text("Another")
    repo.add(".")
    repo.create_commit("Commit on feature_branch")

    monkeypatch.setenv("GIT_REFLOG_ACTION", "merge master")  # Simulate merge
    assert check_is_merge_without_conflict()
    assert not check_is_merge_with_conflict(cwd=tmp_path)


def test_is_merge_with_conflict(tmp_path):
    """
    GIVEN  a merge commit with conflict
    WHEN check_is_merge_with_conflict() or check_is_merge_without_conflict() are called
    THEN they return the expected results
    """
    repo = Repository.create(tmp_path, initial_branch="master")
    repo.create_commit("Initial commit on master")

    repo.create_branch("feature_branch")
    repo.checkout("master")
    conflict_file = tmp_path / "conflict.md"
    conflict_file.write_text("Hello")
    Path(tmp_path / "Other.md").write_text("Other")
    repo.add(".")
    repo.create_commit("Commit on master")

    repo.checkout("feature_branch")
    conflict_file.write_text("World")
    Path(tmp_path / "Another.md").write_text("Another")
    repo.add(".")
    repo.create_commit("Commit on feature_branch")

    # Create merge commit with conflict
    with pytest.raises(subprocess.CalledProcessError) as exc:
        repo.git("merge", "master")

    # check stdout for conflict message
    stdout = exc.value.stdout.decode()
    assert "CONFLICT" in stdout

    assert check_is_merge_with_conflict(cwd=tmp_path)
    assert not check_is_merge_without_conflict()

    # solve conflict but still counts as a merge with conflict
    conflict_file.write_text("Hello World !")
    repo.add(conflict_file)

    assert check_is_merge_with_conflict(cwd=tmp_path)
    assert not check_is_merge_without_conflict()


def test_get_merge_branch_from_reflog(monkeypatch):
    monkeypatch.setenv("GIT_REFLOG_ACTION", "merge master")  # Simulate merge
    assert check_is_merge_without_conflict()
    assert get_merge_branch_from_reflog() == "master"


@my_vcr.use_cassette("test_emoji_filename")
def test_precommit_with_emoji_filename(tmp_path, cli_fs_runner):
    """
    GIVEN a repository with a staged diff on a file with an emoji in its name
    WHEN the precommit command is run
    THEN it executes successfully
    """
    # Set up repository
    repo = Repository.create(tmp_path)

    # Create a file with emoji in the name
    emoji_file = tmp_path / "my_😊_emoji_file.txt"
    emoji_file.write_text("Initial content")
    repo.add(".")
    repo.create_commit("Initial commit with emoji file")

    # Modify the file and stage the changes
    emoji_file.write_text("Modified content")
    repo.add(".")

    # Run the precommit command
    with cd(repo.path):
        result = cli_fs_runner.invoke(cli, ["secret", "scan", "pre-commit"])
    # Verify the command executed successfully
    assert_invoke_ok(result)



import json
from copy import deepcopy
from typing import Any, Dict, TypedDict
from unittest import mock

import pytest
from pygitguardian import GGClient
from pygitguardian.models import ScanResult, SecretIncident
from pytest_voluptuous import S
from voluptuous import Optional as VOptional
from voluptuous import validators

from ggshield.core.config.user_config import SecretConfig
from ggshield.core.scan import Commit
from ggshield.verticals.secret import Result, Results, SecretScanCollection
from ggshield.verticals.secret.output import SecretSARIFOutputHandler
from ggshield.verticals.secret.output.secret_sarif_output_handler import SCHEMA_URL
from ggshield.verticals.secret.secret_scan_collection import Secret
from tests.unit.conftest import (
    _MULTI_SECRET_ONE_LINE_FULL_PATCH,
    _MULTI_SECRET_ONE_LINE_PATCH_SCAN_RESULT,
    _MULTIPLE_SECRETS_PATCH,
    _MULTIPLE_SECRETS_SCAN_RESULT,
    _ONE_LINE_AND_MULTILINE_PATCH,
    _ONE_LINE_AND_MULTILINE_PATCH_SCAN_RESULT,
    SECRET_INCIDENT_MOCK,
)


VERSION_VALIDATOR = validators.Match(r"\d+\.\d+\.\d+")


TOOL_SCHEMA = S(
    {
        "driver": {
            "organization": "GitGuardian",
            "name": "ggshield",
            "informationUri": str,
            "version": VERSION_VALIDATOR,
        },
        "extensions": [
            {
                "name": str,
                "version": VERSION_VALIDATOR,
            }
        ],
    }
)

EMPTY_RESULT_SCHEMA = S(
    {
        "version": "2.1.0",
        "$schema": SCHEMA_URL,
        "runs": [{"tool": TOOL_SCHEMA, "results": []}],
    }
)

MIN_1_INT = validators.Range(min=1)

SARIF_PHYSICAL_LOCATION_DICT_SCHEMA = S(
    {
        "artifactLocation": {
            "uri": str,
        },
        "region": {
            "startLine": MIN_1_INT,
            "startColumn": MIN_1_INT,
            "endLine": MIN_1_INT,
            "endColumn": MIN_1_INT,
        },
    }
)

SARIF_RESULT_DICT_SCHEMA = S(
    {
        "ruleId": str,
        "level": "error",
        "message": {
            "text": str,
            "markdown": str,
        },
        "locations": [
            {
                "physicalLocation": SARIF_PHYSICAL_LOCATION_DICT_SCHEMA,
            }
        ],
        "relatedLocations": [
            {
                "id": int,
                "physicalLocation": SARIF_PHYSICAL_LOCATION_DICT_SCHEMA,
                "message": {"text": str},
            }
        ],
        "partialFingerprints": {"secret/v1": str},
        VOptional("hostedViewerUri"): str,
        VOptional("properties"): {
            VOptional("incidentDetails"): dict,
        },
    }
)

SCHEMA_WITH_INCIDENTS = S(
    {
        "version": "2.1.0",
        "$schema": SCHEMA_URL,
        "runs": [{"tool": TOOL_SCHEMA, "results": [SARIF_RESULT_DICT_SCHEMA]}],
    }
)


@pytest.fixture()
def init_secrets_engine_version():
    # Init secrets engine version: it's not set if we don't make an API call
    with mock.patch(
        "ggshield.verticals.secret.output.secret_sarif_output_handler.VERSIONS"
    ) as versions:
        versions.secrets_engine_version = "3.14.1"
        yield


def test_sarif_output_no_secrets(init_secrets_engine_version):
    """
    GIVEN an empty scan collection
    WHEN SecretSARIFOutputHandler runs on it
    THEN it outputs an empty SARIF document
    """
    scan = SecretScanCollection(id="path", type="test", results=Results())
    handler = SecretSARIFOutputHandler(
        verbose=True, secret_config=SecretConfig(show_secrets=False)
    )
    output = handler._process_scan_impl(scan)
    dct = json.loads(output)

    assert EMPTY_RESULT_SCHEMA == dct


@pytest.mark.parametrize(
    "patch,scan_result",
    [
        pytest.param(
            _MULTIPLE_SECRETS_PATCH,
            _MULTIPLE_SECRETS_SCAN_RESULT,
            id="_MULTIPLE_SECRETS_PATCH",
        ),
        pytest.param(
            _MULTI_SECRET_ONE_LINE_FULL_PATCH,
            _MULTI_SECRET_ONE_LINE_PATCH_SCAN_RESULT,
            id="_MULTI_SECRET_ONE_LINE_FULL_PATCH",
        ),
        pytest.param(
            _ONE_LINE_AND_MULTILINE_PATCH,
            _ONE_LINE_AND_MULTILINE_PATCH_SCAN_RESULT,
            id="_ONE_LINE_AND_MULTILINE_PATCH",
        ),
    ],
)
@pytest.mark.parametrize("with_incident_details", [True, False])
@pytest.mark.parametrize("known_incidents", [True, False])
def test_sarif_output_for_flat_scan_with_secrets(
    init_secrets_engine_version,
    patch: str,
    scan_result: ScanResult,
    with_incident_details: bool,
    known_incidents: bool,
):
    """
    GIVEN a patch containing secrets and a scan result
    WHEN SecretSARIFOutputHandler runs on it
    THEN it outputs a SARIF document pointing to the secrets
    """
    scan_result = deepcopy(scan_result)
    client_mock = mock.Mock(spec=GGClient)
    client_mock.retrieve_secret_incident.return_value = SECRET_INCIDENT_MOCK

    secret_config = SecretConfig(
        show_secrets=False,
        with_incident_details=with_incident_details,
    )
    handler = SecretSARIFOutputHandler(
        verbose=True, secret_config=secret_config, client=client_mock
    )

    commit = Commit.from_patch(patch)
    scannable = next(commit.get_files())

    for index, policy_break in enumerate(scan_result.policy_breaks):
        if known_incidents:
            policy_break.known_secret = True
            policy_break.incident_url = (
                f"https://dashboard.gitguardian.com/workspace/1/incidents/{index}"
            )
        else:
            policy_break.known_secret = False
            policy_break.incident_url = None

    result = Result.from_scan_result(
        file=scannable, scan_result=scan_result, secret_config=SecretConfig()
    )
    results = Results(results=[result])
    scan = SecretScanCollection(id="path", type="test", results=results)

    output = handler._process_scan_impl(scan)
    json_dict = json.loads(output)

    assert SCHEMA_WITH_INCIDENTS == json_dict

    sarif_results = json_dict["runs"][0]["results"]

    # Check each found secret is correctly represented
    for sarif_result, secret in zip(sarif_results, result.secrets):
        check_sarif_result(
            sarif_result,
            scannable.content,
            secret,
            with_incident_details and known_incidents,
        )

    assert len(sarif_results) == len(scan_result.policy_breaks)


PATCHES_AND_RESULTS = [
    (
        _MULTIPLE_SECRETS_PATCH,
        _MULTIPLE_SECRETS_SCAN_RESULT,
    ),
    (
        _MULTI_SECRET_ONE_LINE_FULL_PATCH,
        _MULTI_SECRET_ONE_LINE_PATCH_SCAN_RESULT,
    ),
    (
        _ONE_LINE_AND_MULTILINE_PATCH,
        _ONE_LINE_AND_MULTILINE_PATCH_SCAN_RESULT,
    ),
]


def test_sarif_output_for_nested_scan(init_secrets_engine_version):
    """
    GIVEN a scan results for 3 patches containing secrets
    WHEN SecretSARIFOutputHandler runs on it
    THEN it outputs a SARIF document pointing to the secrets
    """
    handler = SecretSARIFOutputHandler(
        verbose=True, secret_config=SecretConfig(show_secrets=False)
    )

    nested_scans = []
    contents = []
    for idx, (patch, scan_result) in enumerate(PATCHES_AND_RESULTS):
        commit = Commit.from_patch(patch, sha=f"abcd{idx}")
        scannable = next(commit.get_files())
        contents.append(scannable.content)

        result = Result.from_scan_result(
            file=scannable, scan_result=scan_result, secret_config=SecretConfig()
        )
        results = Results(results=[result])
        scan = SecretScanCollection(id=f"nested{idx}", type="test", results=results)
        nested_scans.append(scan)

    scan = SecretScanCollection(id="scan", type="test", scans=nested_scans)

    output = handler._process_scan_impl(scan)
    json_dict = json.loads(output)

    assert SCHEMA_WITH_INCIDENTS == json_dict

    # Create a flat list of secrets
    secrets = sum((s.results.results[0].secrets for s in scan.scans), [])

    # Check each found secret is correctly represented
    sarif_results = json_dict["runs"][0]["results"]
    for content, sarif_result, secret in zip(contents, sarif_results, secrets):
        check_sarif_result(sarif_result, content, secret)

    assert len(sarif_results) == len(secrets)


def check_sarif_result(
    sarif_result: Dict[str, Any],
    content: str,
    secret: Secret,
    contains_incident_details: bool = False,
):
    """Check sarif_result contains a representation of policy_break, applied to content"""

    # Check the secret name
    secret_name = sarif_result["ruleId"]
    assert secret_name == secret.detector

    # Check the matches point to the right part of the content. `expected_matches`
    # and `actual matches` are dicts of match_name => matched_text.
    expected_matches = {
        m.match_type: content[m.index_start : m.index_end + 1] for m in secret.matches
    }

    actual_matches = {}
    for location in sarif_result["relatedLocations"]:
        match_name = location["message"]["text"]
        region = location["physicalLocation"]["region"]
        matched_text = get_content_from_region(content, region)
        actual_matches[match_name] = matched_text

    if contains_incident_details:
        assert (
            "properties" in sarif_result
            and "incidentDetails" in sarif_result["properties"]
        )
        SecretIncident(**sarif_result["properties"]["incidentDetails"])
    else:
        assert "incidentDetails" not in sarif_result

    assert actual_matches == expected_matches


class RegionDict(TypedDict):
    startLine: int
    startColumn: int
    endLine: int
    endColumn: int


def get_content_from_region(content: str, region: RegionDict) -> str:
    # Convert region values into 0-based indices
    # Make end values point *after* the last element
    start_line = region["startLine"] - 1

    # endLine is 1-based but points to the line containing the end, so it does not need
    # to be decreased by 1
    end_line = region["endLine"]

    start_column = region["startColumn"] - 1

    # endColumn is 1-based and points to the character after the match, so it needs to
    # be decreased by 1
    end_column = region["endColumn"] - 1

    lines = content.splitlines()[start_line:end_line]

    # Cut start and end. Do the end first because if we cut the start first then
    # `end_column` will be invalid for 1-line regions
    lines[-1] = lines[-1][:end_column]
    lines[0] = lines[0][start_column:]

    return "\n".join(lines)



from typing import Any, Dict, List, Optional

from marshmallow import fields, post_dump
from pygitguardian.models import Match, MatchSchema

from ggshield.core.filter import censor_string
from ggshield.core.lines import Line
from ggshield.core.match_span import MatchSpan


# The number of lines to display before and after a secret in the patch
NB_CONTEXT_LINES = 3


class ExtendedMatchSchema(MatchSchema):
    pre_line_start = fields.Int(required=False, allow_none=True)
    pre_line_end = fields.Int(required=False, allow_none=True)
    post_line_start = fields.Int(required=False, allow_none=True)
    post_line_end = fields.Int(required=False, allow_none=True)

    @post_dump
    def remove_none_extra(
        self, data: Dict[str, Optional[int]], many: bool
    ) -> Dict[str, Optional[int]]:
        OPTIONAL_OUTPUT_FIELDS = (
            "pre_line_start",
            "pre_line_end",
            "post_line_start",
            "post_line_end",
        )

        return {
            key: value
            for key, value in data.items()
            if key not in OPTIONAL_OUTPUT_FIELDS or value is not None
        }


class ExtendedMatch(Match):
    """Match extended with information about pre and post commit
    line indices"""

    SCHEMA = ExtendedMatchSchema()

    def __init__(
        self,
        span: MatchSpan,
        lines_before_secret: List[Line],
        lines_with_secret: List[Line],
        lines_after_secret: List[Line],
        pre_line_start: Optional[int] = None,
        pre_line_end: Optional[int] = None,
        post_line_start: Optional[int] = None,
        post_line_end: Optional[int] = None,
        **kwargs: Any,
    ):
        self.span = span
        self.lines_before_secret = lines_before_secret
        self.lines_with_secret = lines_with_secret
        self.lines_after_secret = lines_after_secret
        self.pre_line_start = pre_line_start
        self.pre_line_end = pre_line_end
        self.post_line_start = post_line_start
        self.post_line_end = post_line_end
        super().__init__(**kwargs)

    @classmethod
    def from_match(
        cls, match: Match, lines: List[Line], is_patch: bool
    ) -> "ExtendedMatch":
        span = MatchSpan.from_match(match, lines)

        start_line = lines[span.line_index_start]
        end_line = lines[span.line_index_end]
        line_index_start = start_line.pre_index or start_line.post_index
        line_index_end = end_line.pre_index or end_line.post_index
        assert line_index_start is not None and line_index_end is not None

        lines_with_secret = lines[span.line_index_start : span.line_index_end + 1]

        match_split_lines = match.match.splitlines()
        assert len(match_split_lines) == len(lines_with_secret)
        return cls(
            match=match.match,
            index_start=match.index_start,
            index_end=match.index_end,
            span=span,
            lines_before_secret=lines[
                max(
                    span.line_index_start - NB_CONTEXT_LINES + 1, 0
                ) : span.line_index_start
            ],
            lines_with_secret=lines_with_secret,
            lines_after_secret=lines[
                span.line_index_end
                + 1 : min(span.line_index_end + NB_CONTEXT_LINES, len(lines))
            ],
            match_type=match.match_type,
            line_start=line_index_start,
            line_end=line_index_end,
            pre_line_start=start_line.pre_index if is_patch else None,
            post_line_start=start_line.post_index if is_patch else None,
            pre_line_end=end_line.pre_index if is_patch else None,
            post_line_end=end_line.post_index if is_patch else None,
        )

    def censor(self) -> None:
        """
        Censor the match and all the lines containing the secret.
        Lines are modified in place, so the secret does not appear in the context of other secrets.
        """
        len_match = len(self.match)
        self.match = censor_string(self.match)  # Censor the match
        assert len(self.match) == len_match
        match_split_lines = self.match.splitlines()
        assert len(self.lines_with_secret) == len(match_split_lines)
        # Censor the lines containing the secret by replacing the secret by its censored version.
        for index_line, (line, match_split_line) in enumerate(
            zip(self.lines_with_secret, match_split_lines)
        ):
            censor_start = self.span.column_index_start if index_line == 0 else 0
            censor_end = (
                self.span.column_index_end
                if index_line == len(match_split_lines) - 1
                else len(line.content)
            )
            line.content = (  # Modify the line to censor the match
                line.content[:censor_start]
                + match_split_line
                + line.content[censor_end:]
            )

    def __repr__(self) -> str:
        return ", ".join(
            [
                super().__repr__(),
                f"pre_line_start:{self.pre_line_start}",
                f"pre_line_end:{self.pre_line_end}",
                f"post_line_start:{self.post_line_start}",
                f"post_line_end:{self.post_line_end}",
            ]
        )

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, ExtendedMatch):
            return False
        return (
            self.span == other.span
            and self.lines_before_secret == other.lines_before_secret
            and self.lines_with_secret == other.lines_with_secret
            and self.lines_after_secret == other.lines_after_secret
            and self.pre_line_start == other.pre_line_start
            and self.pre_line_end == other.pre_line_end
            and self.post_line_start == other.post_line_start
            and self.post_line_end == other.post_line_end
        )



import os
from pathlib import Path

from platformdirs import user_cache_dir, user_config_dir, user_data_dir

from ggshield.utils.git_shell import NotAGitDirectory, get_git_root


APPNAME = "ggshield"
APPAUTHOR = "GitGuardian"


def get_user_home_dir() -> Path:
    try:
        # See tests/conftest.py for details
        return Path(os.environ["GG_USER_HOME_DIR"])
    except KeyError:
        return Path.home()


def get_config_dir() -> Path:
    try:
        # See tests/conftest.py for details
        return Path(os.environ["GG_CONFIG_DIR"])
    except KeyError:
        return Path(user_config_dir(appname=APPNAME, appauthor=APPAUTHOR))


def get_cache_dir() -> Path:
    try:
        # See tests/conftest.py for details
        return Path(os.environ["GG_CACHE_DIR"])
    except KeyError:
        return Path(user_cache_dir(appname=APPNAME, appauthor=APPAUTHOR))


def get_data_dir() -> Path:
    try:
        # See tests/conftest.py for details
        return Path(os.environ["GG_DATA_DIR"])
    except KeyError:
        return Path(
            user_data_dir(appname=APPNAME, appauthor=APPAUTHOR)
        )  # pragma: no cover


def get_project_root_dir(path: Path) -> Path:
    """
    Returns the source basedir required to find file within filesystem.
    """
    try:
        return get_git_root(wd=path).resolve()
    except NotAGitDirectory:
        # In case we are not in a Git repository
        return path.resolve()



from typing import Optional

import pytest

from ggshield.core.text_utils import clip_long_line, translate_validity


@pytest.mark.parametrize(
    "validity_id, expected",
    [
        ("unknown", "Unknown"),
        (None, "Unknown"),
        ("valid", "Valid"),
        ("unexpected_status", "unexpected_status"),
    ],
)
def test_translate_validity(validity_id: Optional[str], expected: str):
    result = translate_validity(validity_id)
    assert result == expected


@pytest.mark.parametrize(
    "params, want",
    [
        # should not clip anything
        (("123456789", 10, False, False, 0), "123456789"),
        (("123456789", 10, True, False, 0), "123456789"),
        (("123456789", 10, False, True, 0), "123456789"),
        (("123456789", 10, True, True, 0), "123456789"),
        # edge case: should not clip anything
        (("123456789", 9, True, True, 0), "123456789"),
        # edge case: should only clip after
        (("123456789", 8, True, True, 0), "1234567…"),
        # should clip as expected
        (("123456789", 5, False, False, 0), "123456789"),
        (("123456789", 5, True, False, 0), "…6789"),
        (("123456789", 5, False, True, 0), "1234…"),
        (("123456789", 5, True, True, 0), "…456…"),
        (("123456789", 4, True, False, 0), "…789"),
        (("123456789", 4, False, True, 0), "123…"),
        (("123456789", 4, True, True, 0), "…45…"),
        (("123456789", 4, True, True, 6), "…3456…"),
    ],
)
def test_clip_long_line(params, want):
    assert clip_long_line(*params) == want



from pathlib import Path
from subprocess import CalledProcessError

import pytest

from tests.conftest import GG_VALID_TOKEN
from tests.functional.utils import recreate_censored_content, run_ggshield
from tests.repository import Repository


def test_scan_precommit(tmp_path: Path) -> None:
    # GIVEN a repository
    repo = Repository.create(tmp_path)

    # AND ggshield installed as a pre-commit hook
    run_ggshield("install", "-m", "local", "-t", "pre-commit", cwd=repo.path)

    # AND a secret in a file
    secret_file = repo.path / "secret.conf"
    secret_content = f"password = {GG_VALID_TOKEN}"
    secret_file.write_text(secret_content)

    # WHEN I try to commit a secret
    # THEN the hook prevents the commit
    repo.add("secret.conf")
    with pytest.raises(CalledProcessError) as exc:
        repo.create_commit()

    # AND the error message contains the leaked secret
    stderr = exc.value.stderr.decode()
    assert recreate_censored_content(secret_content, GG_VALID_TOKEN) in stderr



from enum import Enum, auto
from pathlib import Path, PurePath, PurePosixPath
from typing import List, Pattern, Set, Union
from urllib.parse import quote

from ggshield.utils._binary_extensions import BINARY_EXTENSIONS
from ggshield.utils.git_shell import (
    get_filepaths_from_ref,
    git_ls,
    git_ls_unstaged,
    is_git_dir,
)


class ListFilesMode(Enum):
    """
    Control `get_filepaths()` behavior:

    - ALL: list all specified paths. If one of the path is a directory, list all its paths recursively.
    - ALL_BUT_GITIGNORED: like ALL, except those ignored by git (listed in .gitignore).
    - GIT_COMMITTED_OR_STAGED: list all committed files and all staged files.
    - GIT_COMMITTED: list only committed files.
    """

    GIT_COMMITTED = auto()
    GIT_COMMITTED_OR_STAGED = auto()
    ALL_BUT_GITIGNORED = auto()
    ALL = auto()


def is_path_excluded(
    path: Union[str, Path], exclusion_regexes: Set[Pattern[str]]
) -> bool:
    path = Path(path)
    if path.is_dir():
        # The directory exclusion regexes have to end with a slash
        # To check if path is excluded, we need to add a trailing slash
        path_string = f"{PurePosixPath(path)}/"
    else:
        path_string = str(PurePosixPath(path))
    return any(r.search(path_string) for r in exclusion_regexes)


def list_files(
    paths: List[Path],
    exclusion_regexes: Set[Pattern[str]],
    list_files_mode: ListFilesMode,
) -> Set[Path]:
    """
    Retrieve a set of the files inside `paths`.

    Note: only plain files are returned, not directories.
    """
    targets: Set[Path] = set()
    for path in paths:
        if path.is_file():
            if (
                list_files_mode == ListFilesMode.ALL_BUT_GITIGNORED
                and is_git_dir(path.parent)
                and path.name not in git_ls_unstaged(path.parent) + git_ls(path.parent)
            ):
                continue
            targets.add(path)
        elif path.is_dir():
            _targets = set()
            if list_files_mode != ListFilesMode.ALL and is_git_dir(path):
                target_filepaths = (
                    get_filepaths_from_ref("HEAD", wd=path)
                    if list_files_mode == ListFilesMode.GIT_COMMITTED
                    else git_ls(path)
                )
                _targets = {path / x for x in target_filepaths}
                if list_files_mode == ListFilesMode.ALL_BUT_GITIGNORED:
                    _targets.update({path / x for x in git_ls_unstaged(path)})
            else:
                _targets = path.rglob(r"*")

            for file_path in _targets:
                if not file_path.is_dir() and not is_path_excluded(
                    file_path, exclusion_regexes
                ):
                    targets.add(file_path)
    return targets


def is_path_binary(path: Union[str, Path]) -> bool:
    ext = Path(path).suffix
    # `[1:]` because `ext` starts with a "." but extensions in `BINARY_EXTENSIONS` do not
    return ext[1:] in BINARY_EXTENSIONS


def url_for_path(path: PurePath) -> str:
    if not path.is_absolute():
        return quote(path.as_posix())

    # Allow ':'. This is required to represent the Windows drive in an URL.
    path_str = quote(path.as_posix(), safe="/:")
    if path_str[0] == "/":
        return f"file://{path_str}"
    else:
        # This happens for Windows paths: `path_str` is something like "c:/foo/bar"
        return f"file:///{path_str}"



from itertools import islice
from typing import Iterable, List, TypeVar


T = TypeVar("T")


def batched(iterable: Iterable[T], batch_size: int) -> Iterable[List[T]]:
    # This function is a backport of the itertools.batched() function  from Python 3.12
    it = iter(iterable)
    while True:
        batch = list(islice(it, batch_size))
        if batch:
            yield batch
        else:
            return



import secrets

import pytest

from ggshield.verticals.hmsl.crypto import (
    DecryptionError,
    decrypt,
    encrypt,
    hash_string,
)


@pytest.fixture
def key():
    return secrets.token_bytes(32)


def test_hash_string():
    string = "test"
    expected_hash = "bdbfe17c6018147f07ce79f8ec415a075d761ccacc14803e7d645155bf21a75f"
    assert hash_string(string) == expected_hash


def test_encrypt_decrypt(key):
    message = "Hello, world!"
    encrypted = encrypt(message, key)
    decrypted = decrypt(encrypted, key)
    assert decrypted == message


def test_encrypt_decrypt_invalid_key(key):
    message = "Hello, world!"
    key2 = secrets.token_bytes(32)

    encrypted = encrypt(message, key)
    with pytest.raises(DecryptionError):
        decrypt(encrypted, key2)


def test_decrypt_invalid_payload(key):
    invalid_payload = b"invalid_payload"

    with pytest.raises(DecryptionError):
        decrypt(invalid_payload, key)



import tempfile
from pathlib import Path
from typing import Any

import click

from ggshield.cmd.secret.scan.secret_scan_common_options import (
    add_secret_scan_common_options,
    create_output_handler,
)
from ggshield.cmd.secret.scan.ui_utils import print_file_list
from ggshield.cmd.utils.context_obj import ContextObj
from ggshield.core import ui
from ggshield.core.client import create_client_from_config
from ggshield.core.errors import UnexpectedError
from ggshield.core.scan import ScanContext, ScanMode
from ggshield.core.scan.file import create_files_from_paths
from ggshield.core.scanner_ui import create_scanner_ui
from ggshield.utils.archive import safe_unpack
from ggshield.utils.click import RealPath
from ggshield.utils.files import ListFilesMode
from ggshield.verticals.secret import SecretScanCollection, SecretScanner


@click.command()
@click.argument(
    "path", nargs=1, type=RealPath(exists=True, resolve_path=True), required=True
)
@add_secret_scan_common_options()
@click.pass_context
def archive_cmd(
    ctx: click.Context,
    path: Path,
    **kwargs: Any,
) -> int:  # pragma: no cover
    """
    Scan an archive file. Supported archive formats are zip, whl, jar, tar, tar.gz, tar.bz2 and tar.xz.
    """
    with tempfile.TemporaryDirectory(suffix="ggshield") as temp_dir:
        temp_path = Path(temp_dir)
        ui.display_heading("Unpacking archive")
        try:
            safe_unpack(path, extract_dir=temp_path)
        except Exception as exn:
            raise UnexpectedError(f'Failed to unpack "{path}" archive: {exn}')

        ctx_obj = ContextObj.get(ctx)
        ctx_obj.client = create_client_from_config(ctx_obj.config)
        config = ctx_obj.config
        files, binary_paths = create_files_from_paths(
            paths=[temp_path],
            exclusion_regexes=ctx_obj.exclusion_regexes,
            list_files_mode=ListFilesMode.ALL,
        )
        print_file_list(files, binary_paths)
        ui.display_heading("Starting scan")

        with create_scanner_ui(len(files)) as scanner_ui:
            scan_context = ScanContext(
                scan_mode=ScanMode.ARCHIVE,
                command_path=ctx.command_path,
            )

            scanner = SecretScanner(
                client=ctx_obj.client,
                cache=ctx_obj.cache,
                scan_context=scan_context,
                secret_config=config.user_config.secret,
            )
            results = scanner.scan(files, scanner_ui=scanner_ui)

        scan = SecretScanCollection(id=path, type="archive_scan", results=results)

        output_handler = create_output_handler(ctx)
        return output_handler.process_scan(scan)



import json
import os
from unittest.mock import patch

import jsonschema
import pytest

from tests.conftest import GG_VALID_TOKEN
from tests.functional.utils import recreate_censored_content, run_ggshield_scan
from tests.repository import Repository


LEAK_CONTENT = f"password = {GG_VALID_TOKEN}"


@pytest.fixture(scope="module")
def leaky_repo(tmp_path_factory: pytest.TempPathFactory) -> Repository:
    """
    Create a repository, add a commit with a secret in it, then add 2 clean commits on
    top of it
    """
    repo = Repository.create(tmp_path_factory.mktemp("test_scan_repo"))

    secret_file = repo.path / "secret.conf"
    secret_file.write_text(LEAK_CONTENT)
    repo.add("secret.conf")
    repo.create_commit()

    for _ in range(2):
        repo.create_commit()

    return repo


def test_scan_repo(leaky_repo: Repository) -> None:
    # GIVEN a repository with a past commit containing a leak
    # WHEN scanning the repo
    # THEN the leak is found
    proc = run_ggshield_scan(
        "repo", str(leaky_repo.path), expected_code=1, cwd=leaky_repo.path
    )

    # AND the output contains the line of the leak
    assert recreate_censored_content(LEAK_CONTENT, GG_VALID_TOKEN) in proc.stdout


def test_scan_repo_json(leaky_repo: Repository, secret_json_schema) -> None:
    # GIVEN a repository with a past commit containing a leak
    # WHEN scanning the repo
    # THEN the leak is found
    proc = run_ggshield_scan(
        "repo", "--json", str(leaky_repo.path), expected_code=1, cwd=leaky_repo.path
    )
    # AND the JSON output matches the expected format
    dct = json.loads(proc.stdout)
    jsonschema.validate(dct, secret_json_schema)


def test_scan_repo_quota_limit_reached(
    leaky_repo: Repository, no_quota_gitguardian_api: str, caplog
) -> None:
    # GIVEN a repository with a past commit containing a leak

    # WHEN scanning the repo
    # THEN error code is 128
    with patch.dict(
        os.environ, {**os.environ, "GITGUARDIAN_API_URL": no_quota_gitguardian_api}
    ):
        proc = run_ggshield_scan(
            "repo",
            str(leaky_repo.path),
            "--json",
            expected_code=128,
            cwd=leaky_repo.path,
        )

    # AND stderr contains an error message
    assert (
        "Error: Could not perform the requested action: no more API calls available."
        in proc.stderr
    )
    # AND stdout is empty
    assert proc.stdout.strip() == ""


def test_scan_repo_exclude_patterns(
    leaky_repo: Repository,
) -> None:
    # GIVEN a repository with a past commit containing a leak
    # AND a pattern to exclude the leak
    # WHEN scanning the repo
    # THEN the leak is not found

    result = run_ggshield_scan(
        "repo",
        "--exclude",
        "secret.conf",
        str(leaky_repo.path),
        expected_code=0,
        cwd=leaky_repo.path,
    )

    assert "No secrets have been found" in result.stdout



import copy
import random
from typing import List, Set

import pytest
from pygitguardian.models import Match, PolicyBreak
from snapshottest import Snapshot

from ggshield.core.filter import censor_match, get_ignore_sha
from tests.unit.conftest import (
    _MULTILINE_SECRET,
    _MULTIPLE_SECRETS_SCAN_RESULT,
    _ONE_LINE_AND_MULTILINE_PATCH_SCAN_RESULT,
    _SIMPLE_SECRET_PATCH_SCAN_RESULT,
)


_FILTERED_MULTILINE_SECRET = """-----BEGIN RSA PRIVATE KEY-----
+MIIBOgIBAAJBAIIRkYjxjE3KIZi******************************+******
+****************************************************************
+****************************************************************
+***********+****************************************************
+****************+***********************************************
+**********************+*****************************************
+****+******Xme/ovcDeM1+3W/UmSHYUW4b3WYq4
+-----END RSA PRIVATE KEY-----"""  # noqa


@pytest.mark.parametrize(
    "policy_breaks, duplicates, expected_shas",
    [
        pytest.param(
            _SIMPLE_SECRET_PATCH_SCAN_RESULT.policy_breaks,
            False,
            {"2b5840babacb6f089ddcce1fe5a56b803f8b1f636c6f44cdbf14b0c77a194c93"},
            id="_SIMPLE_SECRET_PATCH_SCAN_RESULT",
        ),
        pytest.param(
            _SIMPLE_SECRET_PATCH_SCAN_RESULT.policy_breaks,
            True,
            {"2b5840babacb6f089ddcce1fe5a56b803f8b1f636c6f44cdbf14b0c77a194c93"},
            id="_SIMPLE_SECRET_PATCH_SCAN_RESULT-duplicated",
        ),
        pytest.param(
            _MULTIPLE_SECRETS_SCAN_RESULT.policy_breaks,
            False,
            {"41b8889e5e794b21cb1349d8eef1815960bf5257330fd40243a4895f26c2b5c8"},
            id="_MULTIPLE_SECRETS_SCAN_RESULT",
        ),
        pytest.param(
            _MULTIPLE_SECRETS_SCAN_RESULT.policy_breaks,
            True,
            {"41b8889e5e794b21cb1349d8eef1815960bf5257330fd40243a4895f26c2b5c8"},
            id="_MULTIPLE_SECRETS_SCAN_RESULT-duplicated",
        ),
        pytest.param(
            _ONE_LINE_AND_MULTILINE_PATCH_SCAN_RESULT.policy_breaks,
            False,
            {
                "530e5a4a7ea00814db8845dd0cae5efaa4b974a3ce1c76d0384ba715248a5dc1",
                "1945f4a0c42abb19c1a420ddd09b4b4681249a3057c427b95f794b18595e7ffa",
                "060bf63de122848f5efa122fe6cea504aae3b24cea393d887fdefa1529c6a02e",
            },
            id="_MULTIPLE_SECRETS_SCAN_RESULT",
        ),
    ],
)
def test_get_ignore_sha(
    policy_breaks: List[PolicyBreak],
    duplicates: bool,
    expected_shas: Set[str],
    snapshot: Snapshot,
) -> None:
    copy_policy_breaks = copy.deepcopy(policy_breaks)
    if duplicates:
        for policy_break in policy_breaks:
            random.shuffle(policy_break.matches)
        copy_policy_breaks.extend(policy_breaks)

    ignore_shas = {get_ignore_sha(policy_break) for policy_break in copy_policy_breaks}
    if duplicates:
        assert len(ignore_shas) == len(copy_policy_breaks) / 2
    assert ignore_shas == expected_shas


@pytest.mark.parametrize(
    "input_match, expected_value",
    [
        pytest.param(
            Match.SCHEMA.load(
                {
                    "match": "294790898041575",
                    "index_start": 31,
                    "index_end": 46,
                    "type": "client_id",
                }
            ),
            "294*********575",
            id="SIMPLE",
        ),
        pytest.param(
            Match.SCHEMA.load(
                {
                    "match": _MULTILINE_SECRET,
                    "index_start": 31,
                    "index_end": 46,
                    "type": "client_id",
                }
            ),
            _FILTERED_MULTILINE_SECRET,
            id="_MULTILINE_SECRET",
        ),
    ],
)
def test_censor_match(input_match: Match, expected_value: str) -> None:
    value = censor_match(input_match)
    assert len(value) == len(input_match.match)
    assert value == expected_value



import base64
import json
import re
from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import Any, Dict, Iterable, List, Optional, Union, cast

import requests

from ggshield import __version__
from ggshield.utils.itertools import batched

from .crypto import decrypt, make_hint


PREFIX_LENGTH = 5
HASHES_BATCH_SIZE = 100
PREFIXES_BATCH_SIZE = 10
HASH_REGEX = re.compile(r"^[0-9a-f]{64}$")
PREFIX_REGEX = re.compile(f"^[0-9a-f]{{{PREFIX_LENGTH}}}$")


@dataclass
class Secret:
    hash: str
    count: int
    url: Union[str, None] = None


@dataclass
class Match:
    hint: str
    payload: str

    def decrypt(self, hash: str) -> Secret:
        key = bytes.fromhex(hash)
        payload = base64.b64decode(self.payload)
        decrypted = json.loads(decrypt(payload, key))
        if decrypted.get("l"):
            url = decrypted.get("l", {}).get("u")
        else:
            url = None
        return Secret(
            hash=hash,
            count=decrypted["c"],
            url=url,
        )


@dataclass
class SecretsResponse:
    secrets: List[Secret]


@dataclass
class PrefixesResponse:
    matches: List[Match]


@dataclass
class Quota:
    remaining: int
    limit: int
    reset: datetime


class HMSLClient:
    def __init__(
        self,
        url: str,
        hmsl_command_path: str,
        jwt: Optional[str] = None,
        *,
        prefix_length: int = PREFIX_LENGTH,
    ) -> None:
        self.url = url.strip("/")
        self.jwt = jwt
        self.prefix_length = prefix_length
        self._quota: Optional[Quota] = None

        # Create a session with common headers.
        self.session = requests.Session()
        self.session.headers["GGShield-HMSL-Command-Name"] = hmsl_command_path.replace(
            "ggshield ", ""
        ).replace(" ", "_")
        self.session.headers["User-Agent"] = f"GGShield {__version__}"

    @property
    def quota(self) -> Quota:
        """Return the remaining credits."""
        if self._quota is None:
            # Use the side effect of the call to set the remaining credits
            self.check_prefixes([])
        return cast(Quota, self._quota)

    @property
    def status(self) -> bool:
        """Return the status of the HMSL server."""
        try:
            response = self.session.get(f"{self.url}/healthz")
            response.raise_for_status()
            return True
        except requests.exceptions.RequestException:
            return False

    def check(
        self, hashes: Iterable[str], *, full_hashes: bool = False
    ) -> Iterable[Secret]:
        """Check a batch of hashes.
        Supports prefix or full hashes mode.

        Raises a ValueError if a hash is invalid.
        """
        batch_size = HASHES_BATCH_SIZE if full_hashes else PREFIXES_BATCH_SIZE
        for batch in batched(hashes, batch_size):
            if full_hashes:
                yield from self.check_hashes(batch).secrets
            else:
                hints = {make_hint(hash): hash for hash in batch}
                for match in self.check_prefixes(batch).matches:
                    if match.hint in hints:
                        hash = hints[match.hint]
                        yield match.decrypt(hash)

    def query(
        self, hashes: Iterable[str], *, full_hashes: bool = False
    ) -> Iterable[Union[Secret, Match]]:
        """Check a batch of hashes.
        Don't decrypt payloads, return them as is.
        This is mostly useful for the `hmsl query` command,
        clients should use check() instead.
        """
        batch_size = HASHES_BATCH_SIZE if full_hashes else PREFIXES_BATCH_SIZE
        for batch in batched(hashes, batch_size):
            if full_hashes:
                yield from self.check_hashes(batch).secrets
            else:
                yield from self.check_prefixes(batch).matches

    def check_hashes(self, hashes: List[str]) -> SecretsResponse:
        """Audit a batch of full hashes."""
        response = self._query("/v1/hashes", {"hashes": hashes})
        return SecretsResponse(
            secrets=[
                Secret(
                    hash=secret["hash"],
                    count=secret["count"],
                    url=secret["location"]["u"] if secret["location"] else None,
                )
                for secret in response["secrets"]
            ]
        )

    def check_prefixes(self, prefixes_or_hashes: List[str]) -> PrefixesResponse:
        """Audit a batch of prefixes."""
        # Make sure we don't send full hashes
        prefixes = [element[: self.prefix_length] for element in prefixes_or_hashes]
        response = self._query("/v1/prefixes", {"prefixes": prefixes})
        return PrefixesResponse(
            matches=[
                Match(hint=match["hint"], payload=match["payload"])
                for match in response["matches"]
            ]
        )

    def _query(self, endpoint: str, body: Dict[str, List[str]]) -> Any:
        """Send a query to the HMSL service."""
        if self.jwt is not None:
            headers = {"Authorization": f"Bearer {self.jwt}"}
        else:
            headers = {}
        response = self.session.post(
            self.url + endpoint,
            json=body,
            headers=headers,
        )
        response.raise_for_status()
        now = datetime.now().replace(microsecond=0)
        self._quota = Quota(
            remaining=int(response.headers["RateLimit-Remaining"]),
            limit=int(response.headers["RateLimit-Limit"]),
            reset=now + timedelta(seconds=1 + int(response.headers["RateLimit-Reset"])),
        )
        return response.json()



import time
from unittest.mock import MagicMock

import pytest
from pygitguardian.client import GGClient
from pygitguardian.models import JWTResponse

from ggshield.core.config.config import Config
from ggshield.verticals.hmsl.utils import (
    get_token,
    is_token_valid,
    load_token_from_disk,
    remove_token_from_disk,
)


@pytest.fixture
def hmsl_no_env_vars(monkeypatch):
    monkeypatch.delenv("GITGUARDIAN_API_KEY", raising=False)
    monkeypatch.delenv("GITGUARDIAN_SAAS_URL", raising=False)
    monkeypatch.delenv("GITGUARDIAN_SAAS_API_KEY", raising=False)
    monkeypatch.delenv("GITGUARDIAN_HMSL_URL", raising=False)
    monkeypatch.delenv("GITGUARDIAN_HMSL_AUDIENCE", raising=False)


@pytest.fixture
def create_jwt_mock(monkeypatch):
    mock = MagicMock(return_value=JWTResponse(token="dummy_token"))
    monkeypatch.setenv("GITGUARDIAN_SAAS_API_KEY", "dummy_api_key")
    monkeypatch.setattr(GGClient, "create_jwt", mock)
    return mock


def test_hmsl_default_config_values(isolated_fs, hmsl_no_env_vars):
    """
    GIVEN the default config
    WHEN getting HasMySecretLeaked config values
    THEN the values are correctly set
    """
    config = Config()
    assert config.hmsl_url == "https://api.hasmysecretleaked.com"
    assert config.hmsl_audience == "https://api.hasmysecretleaked.com"
    assert config.saas_api_url == "https://api.gitguardian.com"


def test_hmsl_other_env_config_values(isolated_fs, hmsl_no_env_vars, monkeypatch):
    """
    GIVEN a specific environment
    WHEN getting HasMySecretLeaked config values
    THEN the values are correctly set
    """
    # If the user defines a custom HMSL URL
    monkeypatch.setenv(
        "GITGUARDIAN_HMSL_URL", "https://hasmysecretleaked.env.gitguardian.com"
    )
    # They are also expected to set a corresponding GitGuardian instance
    monkeypatch.setenv("GITGUARDIAN_INSTANCE", "https://dashboard.env.gitguardian.com")
    config = Config()
    assert config.hmsl_url == "https://hasmysecretleaked.env.gitguardian.com"
    assert config.hmsl_audience == "https://hasmysecretleaked.env.gitguardian.com"
    assert config.saas_api_url == "https://api.env.gitguardian.com"


def test_hmsl_region_specific_config_values(isolated_fs, hmsl_no_env_vars, monkeypatch):
    """
    GIVEN another SaaS environment
    WHEN getting HasMySecretLeaked config values
    THEN the values are correctly set
    """
    monkeypatch.setenv(
        "GITGUARDIAN_INSTANCE", "https://dashboard.region3.gitguardian.com"
    )
    config = Config()
    assert config.hmsl_url == "https://api.hasmysecretleaked.com"
    assert config.hmsl_audience == "https://api.hasmysecretleaked.com"
    assert config.saas_api_url == "https://api.region3.gitguardian.com"


def test_no_token(isolated_fs, hmsl_no_env_vars):
    """
    GIVEN the default config (not authenticated)
    WHEN getting a token
    THEN we receive nothing
    """
    config = Config()
    token = get_token(config)
    assert token is None


def test_get_token(isolated_fs, create_jwt_mock: MagicMock):
    """
    GIVEN a logged in session to GIM
    WHEN requesting a token
    THEN we receive it
    """
    config = Config()
    assert load_token_from_disk() is None
    token = get_token(config)
    create_jwt_mock.assert_called_once()
    assert token == "dummy_token"


def test_remove_token(isolated_fs, create_jwt_mock: MagicMock):
    """
    GIVEN a logged in session to GIM
    WHEN I'm calling remove_token_from_disk
    THEN the token has been removed
    """
    config = Config()
    assert load_token_from_disk() is None
    token = get_token(config)
    create_jwt_mock.assert_called_once()
    assert token == "dummy_token"
    remove_token_from_disk()
    assert load_token_from_disk() is None


@pytest.mark.parametrize(
    "token",
    [
        "foo",
        (
            "eyJhbGciOiJSUzI1NiIsImtpZCI6Im1haW5fa2V5IiwidHlwIjoiSldUIn0.eyJzdWIiOjEsInRva2VuX2lkIjoi"
            "ZWFkNDg0MGMtN2FkNC00MTdmLWE3ZjItZGY3ZjY1YjQzMDI5IiwiYXVkIjoiaHR0cHM6Ly9oYXNteXNlY3JldGxl"
            "YWtlZDIuZ2l0Z3VhcmRpYW4uY29tIiwiaXNzIjoiaHR0cHM6Ly9kYXNoYm9hcmQuZ2l0Z3VhcmRpYW4uY29tIiwi"
            "aWF0IjoxNjgwMDAwMDAwLCJleHAiOjE2ODAwMDAxMDAsImhtc2xfcXVvdGEiOjEwMDAwMDAwLCJtZW1iZXJfaWQi"
            "OjQyfQ.BYkiuwBVUBjdma3fYfh4c6KKpbkxHc33nUbXrzT4AzETn5hyHMdns5QFyAYKUBPn0A9e4uEuPG59s0xa_"
            "TbfA_xzgCNeRtVtHR8Z464zLXlG6xtYMjsIi22P_NJSCEAzYRg3f1grWARwPUh6_DPLvNnpwL_2loFM89DdD-RQh"
            "hxAloJ5dQs9XEBH3pMku7Z81nlF45kaVOPm_vDsBAvFXETjxpgBRbNPbqoday6MmfueOE1IlGmgso61ObaNTPNt4"
            "tq_t3Wj3V1JL5hBayYQJVJRE3HzDASZ6qPpgy74vwMLBbgSVFsfuimT0ouWlhhkQXhXVaretWByNCtIyIx6vA"
        ),
        (
            "eyJhbGciOiJSUzI1NiIsImtpZCI6Im1haW5fa2V5IiwidHlwIjoiSldUIn0.eyJzdWIiOjEsInRva2VuX2lkIjoiZWF"
            "kNDg0MGMtN2FkNC00MTdmLWE3ZjItZGY3ZjY1YjQzMDI5IiwiYXVkIjoiaHR0cHM6Ly9oYXNteXNlY3JldGxlYWtlZC"
            "5naXRndWFyZGlhbi5jb20iLCJpc3MiOiJodHRwczovL2Rhc2hib2FyZC5naXRndWFyZGlhbi5jb20iLCJpYXQiOjE2N"
            "zAwMDAwMDAsImV4cCI6MTY3NTAwMDAwMCwiaG1zbF9xdW90YSI6MTAwMDAwMDAsIm1lbWJlcl9pZCI6NDJ9.imU1m_z"
            "Ddx8x50KwAKDfvyYke8avWUimairYnzMCT3CsaoYrgrBJt5vsYTkHkm7feLBwgC8rzjoHL_4j8TiqJJ1TF5x1ryxoBY"
            "wPlqnhAVYVWggvRooAmZg7eLYtnygFdrhc4M20bbJM_MB54kPbrqVRcfx3TQjga8IQX4Mb2irjafXXnga_ji5yPtRiA"
            "u7_kn5XtxypoKoUOT9o_E2sBchnmXkrYOV8zXozWfbrbiHymNzSsJM_8uuK9DHGlkUzTo7G3Ek8iRvyd7qdDU1mBEuG"
            "M9f1jAJ_j6jwiEZuT5tCWq7MtA8DZVIDB4wKn2fyZkGbuyhZLpUtg5brpPP1Vg"
        ),
    ],
)
def test_bad_token_validation(token, monkeypatch):
    """
    GIVEN an invalid token
    WHEN validating it
    THEN we receive False
    """
    monkeypatch.setattr(time, "time", lambda: 1680000001)
    assert is_token_valid(token, "https://hasmysecretleaked.gitguardian.com") is False


def test_good_token_validation(monkeypatch):
    """
    GIVEN a valid token
    WHEN validating it
    THEN we receive True
    """
    token = (
        "eyJhbGciOiJSUzI1NiIsImtpZCI6Im1haW5fa2V5IiwidHlwIjoiSldUIn0.eyJzdWIiOjEsInRva2VuX2lkIjoiZWFk"
        "NDg0MGMtN2FkNC00MTdmLWE3ZjItZGY3ZjY1YjQzMDI5IiwiYXVkIjoiaHR0cHM6Ly9oYXNteXNlY3JldGxlYWtlZC5n"
        "aXRndWFyZGlhbi5jb20iLCJpc3MiOiJodHRwczovL2Rhc2hib2FyZC5naXRndWFyZGlhbi5jb20iLCJpYXQiOjE2ODAw"
        "MDAwMDAsImV4cCI6MTY4MDAwMTAwMCwiaG1zbF9xdW90YSI6MTAwMDAwMDAsIm1lbWJlcl9pZCI6NDJ9.bJT7gxGtL2r"
        "mENKSFk7AxyFPesZpo4hIIkyTf30n88BxtPEhPPoX6xZZTBnSGAMjbye4hZ8L0_-7A7hcQAL-0qPv4F6NBB39MiaKG_l"
        "aURz15qG70hwfaeANM-IQILrkK1qr1Ir4hV8EVqXsLWFFKocWnq-fTkBqF2Qq9MkXjKxPv9Yq2nfRMCBPkVmENkP5VfJ"
        "MhiHiVIaDo3Fb9CuQ_BYkU5h5iG8QQWdI2aiM_86mTqpS3PFaJCQY6R2cxVGFEBm7TsxT3RjQ8afWteNs5zXaKz5y7U0"
        "xYC6_0VRyN66IxCL88NM6HeFO3tjHjD9roql853f8StxbcfbUSdCiXg"
    )
    monkeypatch.setattr(time, "time", lambda: 1680000001)
    assert is_token_valid(token, "https://hasmysecretleaked.gitguardian.com")



import logging
from typing import Any

import click

from ggshield.cmd.utils.common_options import add_common_options
from ggshield.cmd.utils.context_obj import ContextObj
from ggshield.core.text_utils import STYLE, format_text
from ggshield.verticals.hmsl import get_client


logger = logging.getLogger(__name__)


@click.command()
@click.pass_context
@add_common_options()
def status_cmd(
    ctx: click.Context,
    **kwargs: Any,
) -> int:
    """
    Make sure the HasMySecretLeaked service is working properly.
    """

    # Get our client
    config = ContextObj.get(ctx).config
    client = get_client(config, ctx.command_path)

    click.echo(
        f"{format_text('API URL:', STYLE['key'])} {client.url}\n"
        f"{format_text('Authenticated:', STYLE['key'])} {str(client.jwt is not None).lower()}\n"
        f"{format_text('Status:', STYLE['key'])} {format_status(client.status)}\n"
    )

    return 0


def format_status(health_check: bool) -> str:
    (color, status) = ("green", "healthy") if health_check else ("red", "unhealthy")
    return format_text(status, {"fg": color})



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Optional

import click
from pygitguardian import GGClient

from ggshield.core.config.user_config import SecretConfig
from ggshield.core.errors import ExitCode
from ggshield.verticals.secret import SecretScanCollection


class SecretOutputHandler(ABC):
    verbose: bool = False
    client: Optional[GGClient] = None
    output: Optional[Path] = None
    use_stderr: bool = False

    def __init__(
        self,
        verbose: bool,
        secret_config: SecretConfig,
        client: Optional[GGClient] = None,
        output: Optional[Path] = None,
    ):
        self.show_secrets = secret_config.show_secrets
        self.verbose = verbose
        self.client = client
        self.output = output
        self.ignore_known_secrets = secret_config.ignore_known_secrets
        self.with_incident_details = secret_config.with_incident_details

    def process_scan(self, scan: SecretScanCollection) -> ExitCode:
        """Process a scan collection, write the report to :attr:`self.output`

        :param scan: The scan collection to process
        :return: The exit code
        """
        text = self._process_scan_impl(scan)
        if self.output:
            self.output.write_text(text)
        else:
            click.echo(text, err=self.use_stderr)
        return self._get_exit_code(scan)

    @abstractmethod
    def _process_scan_impl(self, scan: SecretScanCollection) -> str:
        """Implementation of scan processing,
        called by :meth:`OutputHandler.process_scan`

        Must return a string for the report.

        :param scan: The scan collection to process
        :return: The content
        """
        raise NotImplementedError()

    def _get_exit_code(self, scan: SecretScanCollection) -> ExitCode:
        if scan.total_secrets_count > 0:
            return ExitCode.SCAN_FOUND_PROBLEMS
        return ExitCode.SUCCESS



from ggshield.cmd.secret.scan.docset import generate_files_from_docsets
from tests.unit.conftest import DATA_PATH


DOCSET_EXAMPLE_PATH = DATA_PATH / "docset-example.jsonl"


class TestDocsetScan:
    def test_generate_files_from_docsets(self):
        input_file = open(DOCSET_EXAMPLE_PATH)
        files = list(generate_files_from_docsets(input_file))
        assert {f.filename for f in files} == {
            "https://github.com/owner/repo/issues/1",
            "https://github.com/owner/repo/issues/3",
            "https://github.com/owner/repo/issues/4",
            "https://github.com/owner/repo/issues/4#issuecomment-10069",
            "https://github.com/owner/repo/issues/4#issuecomment-10070",
            "https://github.com/owner/repo/issues/5",
            "https://github.com/owner/repo/issues/6",
            "https://github.com/owner/repo/issues/7",
            "https://github.com/owner/repo/issues/8",
            "https://github.com/owner/repo/issues/8#issuecomment-10071",
        }



from pathlib import Path
from typing import Any, List

import click

from ggshield.cmd.secret.scan.secret_scan_common_options import (
    add_secret_scan_common_options,
    create_output_handler,
)
from ggshield.cmd.secret.scan.ui_utils import print_file_list
from ggshield.cmd.utils.common_decorators import exception_wrapper
from ggshield.cmd.utils.context_obj import ContextObj
from ggshield.cmd.utils.files import check_directory_not_ignored
from ggshield.core import ui
from ggshield.core.client import create_client_from_config
from ggshield.core.scan import ScanContext, ScanMode, Scannable
from ggshield.core.scan.file import create_files_from_paths
from ggshield.core.scanner_ui import create_scanner_ui
from ggshield.utils.click import RealPath
from ggshield.utils.files import ListFilesMode
from ggshield.verticals.secret import SecretScanCollection, SecretScanner


@click.command()
@click.argument("paths", nargs=-1, type=RealPath(exists=True), required=True)
@click.option("--recursive", "-r", is_flag=True, help="Scan directory recursively.")
@click.option("--yes", "-y", is_flag=True, help="Confirm recursive scan.")
@click.option(
    "--use-gitignore", is_flag=True, help="Honor content of .gitignore files."
)
@add_secret_scan_common_options()
@click.pass_context
@exception_wrapper
def path_cmd(
    ctx: click.Context,
    paths: List[Path],
    recursive: bool,
    yes: bool,
    use_gitignore: bool,
    **kwargs: Any,
) -> int:  # pragma: no cover
    """
    Scan files and directories.
    """
    ctx_obj = ContextObj.get(ctx)
    config = ctx_obj.config
    ctx_obj.client = create_client_from_config(config)
    output_handler = create_output_handler(ctx)

    for path in paths:
        check_directory_not_ignored(path, ctx_obj.exclusion_regexes)

    if not recursive:
        if path := next((x for x in paths if x.is_dir()), None):
            raise click.UsageError(
                f"{click.format_filename(path)} is a directory."
                " Use --recursive to scan directories."
            )

    files, binary_paths = create_files_from_paths(
        paths=paths,
        exclusion_regexes=ctx_obj.exclusion_regexes,
        list_files_mode=(
            ListFilesMode.ALL_BUT_GITIGNORED if use_gitignore else ListFilesMode.ALL
        ),
    )
    print_file_list(files, binary_paths)
    if not yes:
        confirm_scan(files)

    if ui.is_verbose():
        ui.display_heading("Starting scan")
    target = paths[0] if len(paths) == 1 else Path.cwd()
    target_path = target if target.is_dir() else target.parent
    with create_scanner_ui(len(files)) as scanner_ui:
        scan_context = ScanContext(
            scan_mode=ScanMode.PATH,
            command_path=ctx.command_path,
            target_path=target_path,
        )

        scanner = SecretScanner(
            client=ctx_obj.client,
            cache=ctx_obj.cache,
            scan_context=scan_context,
            secret_config=config.user_config.secret,
        )
        results = scanner.scan(files, scanner_ui=scanner_ui)
    scan = SecretScanCollection(
        id=" ".join(str(x) for x in paths), type="path_scan", results=results
    )

    return output_handler.process_scan(scan)


def confirm_scan(files: List[Scannable]) -> None:
    count = len(files)
    if count > 1:
        click.confirm(
            f"{count} files will be scanned. Do you want to continue?",
            abort=True,
            err=True,
        )



from pathlib import Path
from unittest.mock import ANY, Mock, patch

import pytest
from click.testing import CliRunner

from ggshield.__main__ import cli
from ggshield.cmd.secret.scan.secret_scan_common_options import (
    IGNORED_DEFAULT_WILDCARDS,
)
from ggshield.core.errors import ExitCode
from ggshield.core.filter import init_exclusion_regexes
from ggshield.core.scan import ScanContext, ScanMode
from ggshield.utils.git_shell import EMPTY_SHA, EMPTY_TREE
from ggshield.utils.os import cd
from tests.repository import Repository
from tests.unit.conftest import assert_invoke_exited_with, assert_invoke_ok


def create_local_repo_with_remote(work_dir: Path) -> Repository:
    remote_repo_path = work_dir / "remote"
    remote_repo = Repository.create(remote_repo_path)
    remote_repo.create_commit("initial commit")

    local_repo_path = work_dir / "local"
    local_repo = Repository.clone(str(remote_repo_path), local_repo_path)

    return local_repo


class TestPrepush:
    @patch("ggshield.cmd.secret.scan.prepush.get_list_commit_SHA")
    def test_prepush_no_commits(self, get_list_mock: Mock, cli_fs_runner: CliRunner):
        """
        GIVEN a prepush range with 0 commits
        WHEN the command is run
        THEN it should return 0 and warn it was unable to get range
        """
        get_list_mock.return_value = []
        result = cli_fs_runner.invoke(
            cli,
            ["-v", "secret", "scan", "pre-push"],
            env={"PRE_COMMIT_FROM_REF": "a" * 40, "PRE_COMMIT_TO_REF": "b" * 40},
        )
        assert_invoke_ok(result)
        assert "Unable to get commit range." in result.output

    @patch("ggshield.cmd.secret.scan.prepush.scan_commit_range")
    def test_prepush_no_commits_stdin(
        self,
        scan_commit_range_mock: Mock,
        tmp_path,
        cli_fs_runner: CliRunner,
    ):
        """
        GIVEN a repository
        AND a new branch pushed from another repository, without any commit
        WHEN the pre-push command is run on the commits from the push
        THEN it scans nothing
        """
        local_repo = create_local_repo_with_remote(tmp_path)

        branch = "topic"
        local_repo.create_branch(branch)

        sha = local_repo.get_top_sha()

        with cd(str(local_repo.path)):
            result = cli_fs_runner.invoke(
                cli,
                ["-v", "secret", "scan", "pre-push", "origin", local_repo.remote_url],
                input=f"refs/heads/master {sha} refs/heads/master {EMPTY_SHA}\n",
            )

        assert_invoke_ok(result)
        scan_commit_range_mock.assert_not_called()
        assert "Unable to get commit range." in result.output

    @patch("ggshield.cmd.secret.scan.prepush.get_list_commit_SHA")
    @patch("ggshield.cmd.secret.scan.prepush.scan_commit_range")
    @patch("ggshield.cmd.secret.scan.prepush.check_git_dir")
    def test_prepush_too_many(
        self,
        check_dir_mock: Mock,
        scan_commit_range_mock: Mock,
        get_list_mock: Mock,
        cli_fs_runner: CliRunner,
    ):
        """
        GIVEN a prepush range with 51 commits
        WHEN the command is run
        THEN it should return 0 warn too many commits for scanning, and scan last 50
        """
        scan_commit_range_mock.return_value = 0
        get_list_mock.return_value = ["a"] * 51
        result = cli_fs_runner.invoke(
            cli,
            ["-v", "secret", "scan", "pre-push"],
            env={"PRE_COMMIT_FROM_REF": "a" * 40, "PRE_COMMIT_TO_REF": "b" * 40},
        )
        assert_invoke_ok(result)
        scan_commit_range_mock.assert_called_once()
        _, kwargs = scan_commit_range_mock.call_args_list[0]
        assert len(kwargs["commit_list"]) == 50
        assert "Too many commits. Scanning last 50" in result.output

    @patch("ggshield.cmd.secret.scan.prepush.scan_commit_range")
    @pytest.mark.parametrize(
        ["local_ref_env_var", "remote_ref_env_var"],
        [
            pytest.param(
                "PRE_COMMIT_SOURCE",
                "PRE_COMMIT_ORIGIN",
                id="old env names",
            ),
            pytest.param(
                "PRE_COMMIT_FROM_REF",
                "PRE_COMMIT_TO_REF",
                id="new env names",
            ),
        ],
    )
    def test_prepush_pre_commit_framework(
        self,
        scan_commit_range_mock: Mock,
        tmp_path,
        cli_fs_runner: CliRunner,
        local_ref_env_var: str,
        remote_ref_env_var: str,
    ):
        """
        GIVEN a prepush range with 20 commits provided by the pre-commit framework
        WHEN the command is run
        THEN it should pass onto scan and return 0
        AND the `exclusion_regexes` argument of the scan_commit_range() call should
        match IGNORED_DEFAULT_WILDCARDS
        """
        local_repo = create_local_repo_with_remote(tmp_path)
        remote_sha = local_repo.get_top_sha()
        shas = [local_repo.create_commit() for _ in range(20)]

        scan_commit_range_mock.return_value = 0

        env = {local_ref_env_var: shas[-1], remote_ref_env_var: remote_sha}

        with cd(str(local_repo.path)):
            result = cli_fs_runner.invoke(
                cli,
                [
                    "-v",
                    "secret",
                    "scan",
                    "pre-push",
                    "origin",
                    "https://example.com/remote",
                ],
                env=env,
            )

        scan_commit_range_mock.assert_called_once_with(
            client=ANY,
            cache=ANY,
            commit_list=shas,
            output_handler=ANY,
            exclusion_regexes=ANY,
            scan_context=ScanContext(
                scan_mode=ScanMode.PRE_PUSH,
                command_path="cli secret scan pre-push",
                target_path=local_repo.path,
            ),
            secret_config=ANY,
        )
        assert_invoke_ok(result)
        assert "Commits to scan: 20" in result.output

        expected_exclusion_regexes = init_exclusion_regexes(IGNORED_DEFAULT_WILDCARDS)
        expected_exclusion_patterns = [r.pattern for r in expected_exclusion_regexes]
        result_exclusion_regexes = scan_commit_range_mock.call_args_list[0][1][
            "exclusion_regexes"
        ]
        result_exclusion_patterns = [r.pattern for r in result_exclusion_regexes]

        assert sorted(result_exclusion_patterns) == sorted(expected_exclusion_patterns)

    @patch("ggshield.cmd.secret.scan.prepush.get_list_commit_SHA")
    @patch("ggshield.cmd.secret.scan.prepush.scan_commit_range")
    @patch("ggshield.cmd.secret.scan.prepush.check_git_dir")
    def test_prepush_stdin_input_empty(
        self,
        check_dir_mock: Mock,
        scan_commit_range_mock: Mock,
        get_list_mock: Mock,
        cli_fs_runner: CliRunner,
    ):
        """
        GIVEN an empty stdin input
        WHEN the command is run
        THEN it should print nothing to scan and return 0
        """

        result = cli_fs_runner.invoke(
            cli,
            [
                "-v",
                "secret",
                "scan",
                "pre-push",
                "origin",
                "https://example.com/remote",
            ],
            input="",
        )
        assert_invoke_ok(result)
        assert "Deletion event or nothing to scan.\n" in result.output

    @patch("ggshield.cmd.secret.scan.prepush.get_list_commit_SHA")
    @patch("ggshield.cmd.secret.scan.prepush.scan_commit_range")
    @patch("ggshield.cmd.secret.scan.prepush.check_git_dir")
    def test_prepush_new_branch_pre_commit_framework(
        self,
        check_dir_mock: Mock,
        scan_commit_range_mock: Mock,
        get_list_mock: Mock,
        cli_fs_runner: CliRunner,
    ):
        """
        GIVEN a target commit of EMPTY_SHA
        WHEN the command is run
        THEN it should warn of new branch and return 0
        """
        scan_commit_range_mock.return_value = 0
        get_list_mock.return_value = ["a" for _ in range(60)]

        result = cli_fs_runner.invoke(
            cli,
            ["-v", "secret", "scan", "pre-push"],
            env={"PRE_COMMIT_FROM_REF": "a" * 40, "PRE_COMMIT_TO_REF": EMPTY_SHA},
        )
        assert_invoke_ok(result)
        get_list_mock.assert_called_once_with(f"{EMPTY_TREE} {'a' * 40}", max_count=51)
        scan_commit_range_mock.assert_called_once()

        assert "New tree event. Scanning last 50 commits" in result.output
        assert "Commits to scan: 50" in result.output

    @patch("ggshield.cmd.secret.scan.prepush.get_list_commit_SHA")
    @patch("ggshield.cmd.secret.scan.prepush.scan_commit_range")
    @patch("ggshield.cmd.secret.scan.prepush.check_git_dir")
    def test_prepush_deletion(
        self,
        check_dir_mock: Mock,
        scan_commit_range_mock: Mock,
        get_list_mock: Mock,
        cli_fs_runner: CliRunner,
    ):
        """
        GIVEN an origin commit of EMPTY_SHA
        WHEN the command is run
        THEN it should warn of new branch and return 0
        """
        scan_commit_range_mock.return_value = 0
        get_list_mock.return_value = ["a" for _ in range(10)]

        result = cli_fs_runner.invoke(
            cli,
            ["-v", "secret", "scan", "pre-push"],
            env={"PRE_COMMIT_FROM_REF": EMPTY_SHA, "PRE_COMMIT_TO_REF": "a" * 40},
        )
        assert_invoke_ok(result)
        assert "Deletion event or nothing to scan.\n" in result.output

    @patch("ggshield.cmd.secret.scan.prepush.scan_commit_range")
    def test_prepush_stdin_input_no_newline(
        self,
        scan_commit_range_mock: Mock,
        tmp_path,
        cli_fs_runner: CliRunner,
    ):
        """
        GIVEN 20 commits through stdin input
        WHEN the command is run
        THEN it should pass onto scan and return 0
        """
        local_repo = create_local_repo_with_remote(tmp_path)
        remote_sha = local_repo.get_top_sha()
        shas = [local_repo.create_commit() for _ in range(20)]

        scan_commit_range_mock.return_value = 0

        with cd(str(local_repo.path)):
            result = cli_fs_runner.invoke(
                cli,
                [
                    "-v",
                    "secret",
                    "scan",
                    "pre-push",
                    "origin",
                    "https://example.com/remote",
                ],
                input=f"refs/heads/main {shas[-1]} refs/heads/main {remote_sha}",
            )
        assert_invoke_ok(result)
        scan_commit_range_mock.assert_called_once()
        assert "Commits to scan: 20" in result.output

    @pytest.mark.parametrize(
        ["called_with_pre_push_args"],
        [
            (True,),
            (False,),
        ],
    )
    @patch("ggshield.cmd.secret.scan.prepush.scan_commit_range")
    def test_prepush_new_branch(
        self,
        scan_commit_range_mock: Mock,
        tmp_path,
        cli_fs_runner: CliRunner,
        called_with_pre_push_args: bool,
    ):
        """
        GIVEN a cloned repository
        AND a local branch with new commits in it
        WHEN the command is run
        THEN it should only scan the new commits
        """
        local_repo = create_local_repo_with_remote(tmp_path)

        branch = "topic"
        local_repo.create_branch(branch)
        shas = [local_repo.create_commit() for _ in range(3)]

        scan_commit_range_mock.return_value = 0

        cmd = ["-v", "secret", "scan", "pre-push"]
        if called_with_pre_push_args:
            cmd.extend(["origin", local_repo.remote_url])
        with cd(str(local_repo.path)):
            result = cli_fs_runner.invoke(
                cli,
                cmd,
                input=f"refs/heads/{branch} {shas[-1]} refs/heads/{branch} {EMPTY_SHA}\n",
            )

        assert_invoke_ok(result)
        scan_commit_range_mock.assert_called_once_with(
            client=ANY,
            cache=ANY,
            commit_list=shas,
            output_handler=ANY,
            exclusion_regexes=ANY,
            scan_context=ANY,
            secret_config=ANY,
        )

    @patch("ggshield.cmd.secret.scan.prepush.scan_commit_range")
    def test_prepush_new_orphan_branch(
        self,
        scan_commit_range_mock: Mock,
        tmp_path,
        cli_fs_runner: CliRunner,
    ):
        """
        GIVEN a cloned repository
        AND an orphan branch with commits in it
        WHEN the command is run
        THEN it should only scan the orphan branch commits
        """
        local_repo = create_local_repo_with_remote(tmp_path)

        branch = "topic"
        local_repo.create_branch(branch, orphan=True)
        shas = [local_repo.create_commit() for _ in range(3)]

        scan_commit_range_mock.return_value = 0

        with cd(str(local_repo.path)):
            result = cli_fs_runner.invoke(
                cli,
                ["-v", "secret", "scan", "pre-push", "origin", local_repo.remote_url],
                input=f"refs/heads/{branch} {shas[-1]} refs/heads/{branch} {EMPTY_SHA}\n",
            )

        assert_invoke_ok(result)
        scan_commit_range_mock.assert_called_once_with(
            client=ANY,
            cache=ANY,
            commit_list=shas,
            output_handler=ANY,
            exclusion_regexes=ANY,
            scan_context=ANY,
            secret_config=ANY,
        )

    @patch("ggshield.cmd.secret.scan.prepush.scan_commit_range")
    def test_remediation_message(
        self,
        scan_commit_range_mock: Mock,
        tmp_path,
        cli_fs_runner: CliRunner,
    ):
        """
        GIVEN some commits
        WHEN the command is run and some secrets are found
        THEN the remediation message is present in the output
        """
        local_repo = create_local_repo_with_remote(tmp_path)
        remote_sha = local_repo.get_top_sha()
        shas = [local_repo.create_commit() for _ in range(20)]

        scan_commit_range_mock.return_value = ExitCode.SCAN_FOUND_PROBLEMS

        with cd(str(local_repo.path)):
            result = cli_fs_runner.invoke(
                cli,
                [
                    "-v",
                    "secret",
                    "scan",
                    "pre-push",
                    "origin",
                    "https://example.com/remote",
                ],
                input=f"refs/heads/main {shas[-1]} refs/heads/main {remote_sha}",
            )
        assert_invoke_exited_with(result, ExitCode.SCAN_FOUND_PROBLEMS)
        scan_commit_range_mock.assert_called_once()
        assert "Commits to scan: 20" in result.output

        assert (
            """
> How to remediate

  Since the secret was detected before the push BUT after the commit, you need to:
  1. rewrite the git history making sure to replace the secret with its reference (e.g. environment variable).
  2. push again.

  To prevent having to rewrite git history in the future, setup ggshield as a pre-commit hook:
    https://docs.gitguardian.com/ggshield-docs/integrations/git-hooks/pre-commit

> [To apply with caution] If you want to bypass ggshield (false positive or other reason), run:
  - if you use the pre-commit framework:

    SKIP=ggshield-push git push"""
            in result.output
        )



"""
Generate a repo with two branches: master and feature_branch.
(all commits contain DEFAULT_FILE_COUNT files, each of size DEFAULT_FILE_SIZE bytes).
There is an initial commit, then:
 - the feature branch is created and one commit is added (possibly containing secrets),
 - we add 3 more commits on the master branch,
 - we install the ggshield pre-commit hook
 - and finally merge master in feature_branch.

 This gives the following commit graph:
 *   17d3a27 (HEAD -> feature_branch) Merge branch 'master' into feature_branch
|\
| * db20e84 (master) Commit master n°3
| * 53a4cd0 Commit master n°2
| * a78f68d Commit master n°1
* | 9f5717c Commit feature branch
|/
* e1db622 Initial commit

This script has an option to create a conflict file in the feature branch (--with_conflict true),
in which case the merge will fail. We catch the error and automatically resolve the conflict by overwriting the file
then commit the change. ggshield is called in pre-commit mode which allow to test performances.
"""

import argparse
import random
import re
import subprocess
import sys
import time
from enum import Enum
from pathlib import Path
from typing import List, Optional

from tests.functional.utils import run_ggshield
from tests.repository import Repository


DEFAULT_FILE_COUNT = 10
DEFAULT_SECRET_COUNT = 5
DEFAULT_FILE_SIZE = 200 * 1024

AVERAGE_LINE_LENGTH = 80
LINE_VARIATION = 20

SECRET_SUFFIX_CHARS = "abcdefghijklmnopqrstuvwxyz0123456789"


class SecretLocation(str, Enum):
    NO_SECRET = "no_secret"
    MASTER_BRANCH = "master_branch"
    FEATURE_BRANCH = "feature_branch"
    CONFLICT_FILE = "conflict_file"


def generate_file(size):
    content = []
    while len(content) < size:
        line_length = AVERAGE_LINE_LENGTH + random.randrange(
            -LINE_VARIATION, LINE_VARIATION
        )
        for _ in range(line_length):
            content.append(chr(random.randrange(32, 127)))
        content.append("\n")
    remove_pwd = re.compile("pwd", re.IGNORECASE)
    return remove_pwd.sub("not", "".join(content))


def generate_secret():
    var = random.choice(["key", "token", "password"])
    suffix = "".join(random.choice(SECRET_SUFFIX_CHARS) for _ in range(20))
    return f"{var} = 'glpat-{suffix}'"


def plant_secret(path, secret):

    lines = path.read_text().splitlines()

    idx = random.randrange(len(lines))
    lines.insert(idx, secret)

    path.write_text("\n".join(lines))


def generate_commit(
    repo,
    root_dir,
    file_size,
    nb_files,
    nb_secrets,
    commit_message,
    file_prefix="file",
    files_with_conflict: Optional[List[str]] = None,
):
    content = generate_file(file_size)
    for idx in range(nb_files):
        Path(root_dir / f"{file_prefix}-{random.randrange(10000)}").write_text(content)

    # Plant secrets
    files = list(Path(root_dir).glob(f"{file_prefix}-*"))
    for _ in range(nb_secrets):
        while True:
            path = random.choice(files)
            if not path.is_dir():
                break
        secret = generate_secret()
        plant_secret(path, secret)
        print(f"Planted secret in {path}")

    if files_with_conflict is not None:
        for file in files_with_conflict:
            Path(root_dir / file).write_text("conflict")

    repo.add(".")
    repo.create_commit(commit_message)


def generate_repo_with_merge_commit(
    root_dir: Path,
    file_size: int = DEFAULT_FILE_SIZE,
    nb_files_per_commit: int = DEFAULT_FILE_COUNT,
    with_conflict: bool = False,
    secret_location: SecretLocation = SecretLocation.NO_SECRET,
    scan_all_merge_files: bool = False,
) -> None:
    Path(root_dir).mkdir(parents=True, exist_ok=True)
    repo = Repository.create(root_dir, initial_branch="master")
    # First commit, on master no secrets
    generate_commit(repo, root_dir, file_size, nb_files_per_commit, 0, "Initial commit")
    # Create a feature branch
    repo.create_branch("feature_branch")
    repo.checkout("master")
    # 3 more commits, on master no secrets
    for i in range(2):
        generate_commit(
            repo,
            root_dir,
            file_size,
            nb_files_per_commit,
            0,
            f"Commit master n°{i+1}",
            file_prefix=f"file-master-{i+1}",
        )
    # Pick one of the files of HEAD  commit at random
    files_last_commit = repo.git(*["diff", "--name-only", "HEAD~1"]).splitlines()
    generate_commit(
        repo,
        root_dir,
        file_size,
        nb_files_per_commit,
        DEFAULT_SECRET_COUNT if secret_location == SecretLocation.MASTER_BRANCH else 0,
        f"Commit master n°{3}",
        file_prefix=f"file-master-{3}",
    )

    repo.checkout("feature_branch")
    # Additional commit, on feature_branch with secrets
    files_with_conflict = [random.choice(files_last_commit)] if with_conflict else []
    print(f"Files with conflict: {files_with_conflict}")
    generate_commit(
        repo,
        root_dir,
        file_size,
        nb_files_per_commit,
        DEFAULT_SECRET_COUNT if secret_location == SecretLocation.MASTER_BRANCH else 0,
        "Commit feature branch",
        files_with_conflict=files_with_conflict,
        file_prefix="file-feature",
    )

    # Install ggshield hook
    run_ggshield(
        "install",
        "-m",
        "local",
        "-t",
        "pre-commit",
        cwd=repo.path,
    )
    if scan_all_merge_files:
        # rewrite the git hook file to add the option --skip-unchanged-merge-files
        hook_path = Path(
            f"{root_dir}/.git/hooks/pre-commit",
        )
        with open(hook_path, "r") as f:
            hook = f.read()
        hook = hook.replace(r"pre-commit", r"pre-commit --scan-all-merge-files")
        with open(hook_path, "w") as f:
            f.write(hook)

    if not with_conflict:
        # Create merge commit
        start = time.perf_counter()
        repo.git("merge", "master")
        end = time.perf_counter()
        print(
            f"GGshield scan on merge commit (no conflict) with ggshield took: {end - start:.6f} seconds"
        )
    else:
        # Create merge commit with conflict
        try:
            repo.git("merge", "master")
        except subprocess.CalledProcessError:
            for file in files_with_conflict:
                Path(root_dir / file).write_text(
                    generate_secret()
                    if secret_location == SecretLocation.CONFLICT_FILE
                    else "conflict solved !"
                )
            repo.add(".")
            start = time.perf_counter()
            repo.create_commit("Solved conflict")
            end = time.perf_counter()
            print(
                "GGshield scan on merge commit with conflict took: {:.6f} seconds".format(
                    end - start
                )
            )


def main():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter, description=__doc__
    )

    parser.add_argument(
        "-s", "--size", type=int, help="file size", default=DEFAULT_FILE_SIZE
    )

    parser.add_argument(
        "-c", "--count", type=int, help="number of files", default=DEFAULT_FILE_COUNT
    )

    parser.add_argument(
        "--with_conflict",
        type=bool,
        help="Whether to create a conflict file",
        default=False,
    )

    parser.add_argument(
        "--secret_location",
        type=SecretLocation,
        help="where to add secrets",
        default=SecretLocation.NO_SECRET,
    )

    parser.add_argument("dir", help="where to generate the files")

    args = parser.parse_args()
    root_dir = Path(args.dir)
    generate_repo_with_merge_commit(
        root_dir,
        args.size,
        args.count,
        args.with_conflict,
        args.secret_location,
    )

    return 0


if __name__ == "__main__":
    sys.exit(main())



import os

from ggshield.core import ui


def check_user_requested_skip() -> bool:
    if get_skip_env_var():
        ui.display_warning("Skipping ggshield hook based on SKIP environment variable.")
        return True
    return False


def get_skip_env_var() -> bool:
    """
    When `ggshield` is configured via the [pre-commit
    framework](https://pre-commit.com/) the user has the option to skip
    specific scans by setting the `SKIP` environment variable. When `ggshield`
    is configured directly as a `git` pre-commit hook, though, the user only
    has the option to disable _all_ pre-commit hooks via the `--no-verify`
    command-line option. Here we're "borrowing" the pre-commit framework's
    option so ggshield can be skipped the same way regardless of how it's
    configured.
    """
    skip_env_var = os.getenv("SKIP")
    if skip_env_var is None:
        return False

    return any(p.strip().lower() == "ggshield" for p in skip_env_var.split(","))



from typing import Dict, Optional
from unittest.mock import Mock, patch

import pytest

from ggshield.core.env_utils import TRACKED_ENV_VARS, load_dot_env
from ggshield.utils.os import cd


@patch("ggshield.core.env_utils.load_dotenv")
@pytest.mark.parametrize(
    ["cwd", "env", "expected_dotenv_path"],
    [
        ("sub", {}, None),
        (".", {}, ".env"),
        (".", {"GITGUARDIAN_DONT_LOAD_ENV": "1"}, None),
        (".", {"GITGUARDIAN_DOTENV_PATH": ".custom-env"}, ".custom-env"),
        (".", {"GITGUARDIAN_DOTENV_PATH": ".does-not-exist"}, None),
    ],
)
def test_load_dot_env(
    load_dotenv_mock: Mock,
    monkeypatch,
    tmp_path,
    cwd: str,
    env: Dict[str, str],
    expected_dotenv_path: Optional[str],
):
    """
    GIVEN a file hierarchy like this:
    /sub/
    /.env
    /.custom-env
    AND environment variables set according to `env`
    AND the current working directory being `/sub/`
    WHEN load_dot_env() is called
    THEN the appropriate environment file is loaded
    """
    (tmp_path / "sub").mkdir()
    (tmp_path / ".env").touch()
    (tmp_path / ".custom-env").touch()

    if env:
        for key, value in env.items():
            monkeypatch.setenv(key, value)

    with cd(tmp_path / cwd):
        load_dot_env()

        if expected_dotenv_path:
            expected_dotenv_path = tmp_path / expected_dotenv_path
            load_dotenv_mock.assert_called_once_with(
                expected_dotenv_path, override=True
            )
        else:
            load_dotenv_mock.assert_not_called()


@patch("ggshield.core.env_utils.get_git_root")
@patch("ggshield.core.env_utils.is_git_dir")
@patch("ggshield.core.env_utils.load_dotenv")
def test_load_dot_env_loads_git_root_env(
    load_dotenv_mock: Mock, is_git_dir_mock, get_git_root_mock, tmp_path
):
    """
    GIVEN a git repository checkout with this file hierarchy:
    /sub1/sub2
    /sub1/.env
    /.env
    AND the current working directory being `/sub1/sub2`
    WHEN load_dot_env() is called
    THEN the .env file at the root of the git repository is loaded
    """
    is_git_dir_mock.return_value = True
    get_git_root_mock.return_value = tmp_path

    sub1_sub2_dir = tmp_path / "sub1" / "sub2"
    git_root_dotenv = tmp_path / ".env"

    sub1_sub2_dir.mkdir(parents=True)
    (tmp_path / "sub1" / ".env").touch()
    git_root_dotenv.touch()

    with cd(sub1_sub2_dir):
        load_dot_env()
        load_dotenv_mock.assert_called_once_with(git_root_dotenv, override=True)


@pytest.mark.parametrize("env_var", sorted(TRACKED_ENV_VARS))
def test_load_dot_env_returns_set_vars(env_var, tmp_path, monkeypatch):
    """
    GIVEN an env var that is set, and also set with the same value in the .env
    WHEN load_dot_env() is called
    THEN it returns the env var
    """
    monkeypatch.setenv(env_var, "value")
    (tmp_path / ".env").write_text(f"{env_var}=value")
    with cd(tmp_path):
        set_variables = load_dot_env()

    assert set_variables == {env_var}



from pathlib import Path
from typing import Optional, Pattern, Set, cast

import click
from pygitguardian import GGClient

from ggshield.cmd.utils.output_format import OutputFormat
from ggshield.core.cache import Cache
from ggshield.core.config import Config


class ContextObj:
    """This object is used as Click "context object": the object accessible via
    `ctx.obj`.

    While it is possible to use `ctx.obj` directly, it's better to write:

    ```
    ctx_obj = ContextObj.get(ctx)
    ```

    Doing so ensures the type checker knows `ctx_obj` is an instance of
    ContextObj and can verify access to member variables, whereas when using
    `ctx.obj` directly, the type checker considers it to be of type `Any`.

    Some attributes (`config`, `client` and `cache`) are implemented as
    properties. These properties check the attribute has been set before
    returning it. This allows the class to store instance `foo` as
    `Optional[Foo]` but exposing it as `Foo`.
    """

    def __init__(self):
        self._config: Optional[Config] = None
        self._client: Optional[GGClient] = None
        self._cache: Optional[Cache] = None

        # Depending on the vertical, this is set by configuration options or
        # command-line parameters
        self.exclusion_regexes: Set[Pattern[str]] = set()

        # Set to false by the --no-check-for-updates option
        self.check_for_updates = True

        self.output_format = OutputFormat.TEXT

        # Set by the --output option
        self.output: Optional[Path] = None

    @property
    def use_json(self) -> bool:
        return self.output_format == OutputFormat.JSON

    @property
    def config(self) -> Config:
        assert self._config
        return self._config

    @config.setter
    def config(self, value: Config) -> None:
        self._config = value

    @property
    def client(self) -> GGClient:
        assert self._client
        return self._client

    @client.setter
    def client(self, value: GGClient) -> None:
        self._client = value

    @property
    def cache(self) -> Cache:
        assert self._cache
        return self._cache

    @cache.setter
    def cache(self, value: Cache) -> None:
        self._cache = value

    @staticmethod
    def get(ctx: click.Context) -> "ContextObj":
        """The recommended way to get a ContextObj instance, see the class docstring for
        details"""
        assert ctx.obj
        return cast(ContextObj, ctx.obj)



from ggshield.utils.git_shell import check_git_dir

from ... import ui
from .commit_range import collect_commit_range_from_ci_env
from .previous_commit import get_previous_commit_from_ci_env


def get_current_and_previous_state_from_ci_env():
    """
    Returns the current commit sha and the previous commit sha of the targeted
    branch in a CI env
    """
    check_git_dir()

    new_commits, _ = collect_commit_range_from_ci_env()
    previous_commit = get_previous_commit_from_ci_env()

    if not new_commits:
        current_commit = "HEAD"
    else:
        current_commit = new_commits[-1]

    if new_commits:
        ui.display_verbose("List of new commits: ")
        for commit in new_commits:
            ui.display_verbose(f"- {commit}")

    ui.display_verbose(
        f"Comparing commit {current_commit} to commit {previous_commit}",
    )

    return current_commit, previous_commit



import logging
import re
import time
from typing import Optional

import jwt
from pygitguardian.models import Detail, JWTService

from ggshield.core.client import create_client
from ggshield.core.config.config import Config
from ggshield.core.dirs import get_cache_dir
from ggshield.core.errors import (
    AuthExpiredError,
    MissingTokenError,
    UnknownInstanceError,
)
from ggshield.verticals.hmsl.client import HMSLClient


logger = logging.getLogger(__name__)


# Tools for parsing env files
ENV_LINE_REGEX = re.compile(r'(\S+)(?: *?)=(?: *?)((?:".*?[^\\]")|\S+)')

EXCLUDED_KEYS = {
    "HOST",
    "HOSTNAME",
    "PATH",
    "PORT",
}

EXCLUDED_VALUES = {
    "0",
    "1",
    "disabled",
    "enabled",
    "false",
    "n",
    "no",
    "none",
    "null",
    "off",
    "on",
    "true",
    "y",
    "yes",
}


def get_client(config: Config, hmsl_command_path: str) -> HMSLClient:
    """
    Get an instance of HMSLClient according to your current CLI config and the
    current CLI command.
    """
    token = get_token(config)
    return HMSLClient(
        url=config.hmsl_url, hmsl_command_path=hmsl_command_path, jwt=token
    )


def get_token_path():
    return get_cache_dir() / "hmsl_token"


def get_token(config: Config) -> Optional[str]:
    """Get a JWT token to use the HMSL service.
    If we are not logged, no token is returned and
    the client uses the HMSL service anonymously.
    """

    # Look for a stored token
    token = load_token_from_disk()

    if is_token_valid(token, config.hmsl_audience):
        logger.debug("Using cached HMSL token")
        return token

    logger.debug("No valid token cached. Getting a new one.")

    try:
        client = create_client(
            api_url=config.saas_api_url,
            api_key=config.saas_api_key,
            allow_self_signed=config.user_config.allow_self_signed,
        )
        audience = config.hmsl_audience
    except (MissingTokenError, AuthExpiredError):
        logger.debug("No API key found, using HMSL anonymously.")
        return None
    except UnknownInstanceError as e:
        logger.warning(f"Unknown instance {e.instance}, using HMSL anonymously.")
        return None
    except Exception as e:
        logger.warning(f"Unexpected error: {e}. Using HMSL anonymously.")
        return None

    # Get a new token
    logger.debug("Requesting new JWT token")
    response = client.create_jwt(audience, JWTService.HMSL)
    if isinstance(response, Detail):
        logger.warning(f"Unexpected error: {response.detail}. Using HMSL anonymously.")
        return None
    token = response.token

    # Cache it for future calls and return it
    save_token(token)
    return token


def is_token_valid(token: Optional[str], audience: str) -> bool:
    if not token:
        return False
    try:
        # We only check expiration at this point.
        decoded = jwt.decode(
            token, options={"verify_signature": False, "require": ["exp", "aud"]}
        )
        # If we changed the target audience (only useful during tests)
        if decoded["aud"] != audience:
            return False
        # Keep one minute of leeway
        return int(decoded["exp"]) > time.time() + 60
    except Exception:
        pass
    return False


def load_token_from_disk() -> Optional[str]:
    try:
        return get_token_path().read_text()
    except FileNotFoundError:
        return None
    except Exception as e:
        logger.warning(f"Error while loading token: {e}")
    return None


def save_token(token: str) -> None:
    try:
        get_token_path().write_text(token)
    except Exception as e:
        logger.warning(f"Error while saving token: {e}")


def remove_token_from_disk() -> None:
    try:
        get_token_path().unlink(missing_ok=True)
    except OSError:
        pass
    except Exception as e:
        logger.warning(f"Error while removing token: {e}")
